from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ...arctrl_core.data_file import DataFile
from ...arctrl_json.data_file import (ISAJson_decoder, ISAJson_encoder, ROCrate_decoder, ROCrate_encoder)
from ...arctrl_json.encode import default_spaces
from ...fable_library.result import FSharpResult_2
from ...fable_library.string_ import (to_text, printf)
from ...thoth_json_core.types import IEncodable
from ...thoth_json_python.decode import Decode_fromString
from ...thoth_json_python.encode import to_string

def ARCtrl_DataFile__DataFile_fromISAJsonString_Static_Z721C83C5(s: str) -> DataFile:
    match_value: FSharpResult_2[DataFile, str] = Decode_fromString(ISAJson_decoder, s)
    if match_value.tag == 1:
        raise Exception(to_text(printf("Error decoding string: %O"))(match_value.fields[0]))

    else: 
        return match_value.fields[0]



def ARCtrl_DataFile__DataFile_toISAJsonString_Static_71136F3F(spaces: int | None=None) -> Callable[[DataFile], str]:
    def _arrow3792(f: DataFile, spaces: Any=spaces) -> str:
        value: IEncodable = ISAJson_encoder(f)
        return to_string(default_spaces(spaces), value)

    return _arrow3792


def ARCtrl_DataFile__DataFile_ToISAJsonString_71136F3F(this: DataFile, spaces: int | None=None) -> str:
    return ARCtrl_DataFile__DataFile_toISAJsonString_Static_71136F3F(spaces)(this)


def ARCtrl_DataFile__DataFile_fromROCrateJsonString_Static_Z721C83C5(s: str) -> DataFile:
    match_value: FSharpResult_2[DataFile, str] = Decode_fromString(ROCrate_decoder, s)
    if match_value.tag == 1:
        raise Exception(to_text(printf("Error decoding string: %O"))(match_value.fields[0]))

    else: 
        return match_value.fields[0]



def ARCtrl_DataFile__DataFile_toROCrateJsonString_Static_71136F3F(spaces: int | None=None) -> Callable[[DataFile], str]:
    def _arrow3793(f: DataFile, spaces: Any=spaces) -> str:
        value: IEncodable = ROCrate_encoder(f)
        return to_string(default_spaces(spaces), value)

    return _arrow3793


def ARCtrl_DataFile__DataFile_ToROCrateJsonString_71136F3F(this: DataFile, spaces: int | None=None) -> str:
    return ARCtrl_DataFile__DataFile_toROCrateJsonString_Static_71136F3F(spaces)(this)


__all__ = ["ARCtrl_DataFile__DataFile_fromISAJsonString_Static_Z721C83C5", "ARCtrl_DataFile__DataFile_toISAJsonString_Static_71136F3F", "ARCtrl_DataFile__DataFile_ToISAJsonString_71136F3F", "ARCtrl_DataFile__DataFile_fromROCrateJsonString_Static_Z721C83C5", "ARCtrl_DataFile__DataFile_toROCrateJsonString_Static_71136F3F", "ARCtrl_DataFile__DataFile_ToROCrateJsonString_71136F3F"]




from __future__ import annotations
from dataclasses import dataclass
from typing import Any
from ...fable_library.list import (length, empty, FSharpList, choose, append)
from ...fable_library.option import default_arg
from ...fable_library.reflection import (TypeInfo, string_type, option_type, list_type, record_type)
from ...fable_library.string_ import (to_text, printf)
from ...fable_library.types import (to_string, Record)
from ..ontology_annotation import OntologyAnnotation
from .factor_value import (FactorValue, FactorValue_reflection)
from .material_attribute_value import (MaterialAttributeValue, MaterialAttributeValue_reflection)
from .source import (Source, Source_reflection)

def _expr760() -> TypeInfo:
    return record_type("ARCtrl.Process.Sample", [], Sample, lambda: [("ID", option_type(string_type)), ("Name", option_type(string_type)), ("Characteristics", option_type(list_type(MaterialAttributeValue_reflection()))), ("FactorValues", option_type(list_type(FactorValue_reflection()))), ("DerivesFrom", option_type(list_type(Source_reflection())))])


@dataclass(eq = False, repr = False, slots = True)
class Sample(Record):
    ID: str | None
    Name: str | None
    Characteristics: FSharpList[MaterialAttributeValue] | None
    FactorValues: FSharpList[FactorValue] | None
    DerivesFrom: FSharpList[Source] | None
    def Print(self, __unit: None=None) -> str:
        this: Sample = self
        return to_string(this)

    def PrintCompact(self, __unit: None=None) -> str:
        this: Sample = self
        chars: int = length(default_arg(this.Characteristics, empty())) or 0
        facts: int = length(default_arg(this.FactorValues, empty())) or 0
        arg: str = Sample__get_NameAsString(this)
        return to_text(printf("%s [%i characteristics; %i factors]"))(arg)(chars)(facts)


Sample_reflection = _expr760

def Sample_make(id: str | None=None, name: str | None=None, characteristics: FSharpList[MaterialAttributeValue] | None=None, factor_values: FSharpList[FactorValue] | None=None, derives_from: FSharpList[Source] | None=None) -> Sample:
    return Sample(id, name, characteristics, factor_values, derives_from)


def Sample_create_62424AD2(Id: str | None=None, Name: str | None=None, Characteristics: FSharpList[MaterialAttributeValue] | None=None, FactorValues: FSharpList[FactorValue] | None=None, DerivesFrom: FSharpList[Source] | None=None) -> Sample:
    return Sample_make(Id, Name, Characteristics, FactorValues, DerivesFrom)


def Sample_get_empty(__unit: None=None) -> Sample:
    return Sample_create_62424AD2()


def Sample__get_NameAsString(this: Sample) -> str:
    return default_arg(this.Name, "")


def Sample_getCharacteristicUnits_53716792(s: Sample) -> FSharpList[OntologyAnnotation]:
    def chooser(c: MaterialAttributeValue, s: Any=s) -> OntologyAnnotation | None:
        return c.Unit

    return choose(chooser, default_arg(s.Characteristics, empty()))


def Sample_getFactorUnits_53716792(s: Sample) -> FSharpList[OntologyAnnotation]:
    def chooser(c: FactorValue, s: Any=s) -> OntologyAnnotation | None:
        return c.Unit

    return choose(chooser, default_arg(s.FactorValues, empty()))


def Sample_getUnits_53716792(s: Sample) -> FSharpList[OntologyAnnotation]:
    return append(Sample_getCharacteristicUnits_53716792(s), Sample_getFactorUnits_53716792(s))


def Sample_setCharacteristicValues(values: FSharpList[MaterialAttributeValue], s: Sample) -> Sample:
    return Sample(s.ID, s.Name, values, s.FactorValues, s.DerivesFrom)


def Sample_setFactorValues(values: FSharpList[FactorValue], s: Sample) -> Sample:
    return Sample(s.ID, s.Name, s.Characteristics, values, s.DerivesFrom)


__all__ = ["Sample_reflection", "Sample_make", "Sample_create_62424AD2", "Sample_get_empty", "Sample__get_NameAsString", "Sample_getCharacteristicUnits_53716792", "Sample_getFactorUnits_53716792", "Sample_getUnits_53716792", "Sample_setCharacteristicValues", "Sample_setFactorValues"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ..fable_library.list import (is_empty, tail, FSharpList, head, cons, take_while as take_while_1, skip_while, reverse, empty, of_array, of_seq)
from ..fable_library.option import value as value_1
from ..fable_library.seq import (length, take_while)
from ..fable_library.system_text import (StringBuilder__ctor, StringBuilder__AppendLine_Z721C83C5)
from ..fable_library.types import to_string
from ..fable_library.util import (get_enumerator, ignore)
from .persil import pipeline
from .yamlicious_types import (Config, Config_init_71136F3F, Config__get_WhitespaceString, PreprocessorElement, Preprocessor)

def ReadHelpers_indentLevel(line: str) -> int:
    def predicate(c: str, line: Any=line) -> bool:
        return c == " "

    return length(take_while(predicate, line))


def write(root_element: PreprocessorElement, fconfig: Callable[[Config], Config] | None=None) -> str:
    config_1: Config
    config: Config = Config_init_71136F3F()
    config_1 = value_1(fconfig)(config) if (fconfig is not None) else config
    sb: Any = StringBuilder__ctor()
    def loop(current: PreprocessorElement, sb_1: Any, config_2: Config, root_element: Any=root_element, fconfig: Any=fconfig) -> None:
        if current.tag == 1:
            next_config: Config = Config(config_2.Whitespace, config_2.Level + 1)
            with get_enumerator(current.fields[0]) as enumerator:
                while enumerator.System_Collections_IEnumerator_MoveNext():
                    loop(enumerator.System_Collections_Generic_IEnumerator_1_get_Current(), sb_1, next_config)

        elif current.tag == 0:
            with get_enumerator(current.fields[0]) as enumerator_1:
                while enumerator_1.System_Collections_IEnumerator_MoveNext():
                    loop(enumerator_1.System_Collections_Generic_IEnumerator_1_get_Current(), sb_1, config_2)

        elif current.tag == 3:
            pass

        else: 
            ignore(StringBuilder__AppendLine_Z721C83C5(sb_1, Config__get_WhitespaceString(config_2) + current.fields[0]))


    loop(root_element, sb, config_1)
    return to_string(sb)


def read(yaml_str: str) -> Preprocessor:
    content: dict[str, Any] = pipeline(yaml_str)
    def loop(lines_mut: FSharpList[str], current_intendation_mut: int, acc_mut: FSharpList[PreprocessorElement], yaml_str: Any=yaml_str) -> FSharpList[PreprocessorElement]:
        while True:
            (lines, current_intendation, acc) = (lines_mut, current_intendation_mut, acc_mut)
            if not is_empty(lines):
                rest: FSharpList[str] = tail(lines)
                line: str = head(lines)
                line_ele: PreprocessorElement = PreprocessorElement(2, line.strip())
                next_intendation: int = ReadHelpers_indentLevel(line) or 0
                if next_intendation == current_intendation:
                    lines_mut = rest
                    current_intendation_mut = current_intendation
                    acc_mut = cons(PreprocessorElement(2, line.strip()), acc)
                    continue

                else: 
                    def predicate(l: str, lines: Any=lines, current_intendation: Any=current_intendation, acc: Any=acc) -> bool:
                        return ReadHelpers_indentLevel(l) > current_intendation

                    next_level_lines: FSharpList[str] = take_while_1(predicate, rest)
                    def predicate_1(l_1: str, lines: Any=lines, current_intendation: Any=current_intendation, acc: Any=acc) -> bool:
                        return ReadHelpers_indentLevel(l_1) > current_intendation

                    lines_mut = skip_while(predicate_1, rest)
                    current_intendation_mut = current_intendation
                    acc_mut = cons(PreprocessorElement(1, cons(line_ele, reverse(loop(next_level_lines, next_intendation, empty())))), acc)
                    continue


            else: 
                return acc

            break

    return Preprocessor(PreprocessorElement(0, reverse(loop(of_array(content["Lines"]), 0, empty()))), content["StringMap"], content["CommentMap"])


def mk_line(line: str) -> PreprocessorElement:
    return PreprocessorElement(2, line)


def mkl_level(children: Any | None=None) -> PreprocessorElement:
    return PreprocessorElement(0, of_seq(children))


def mk_intendation(children: Any | None=None) -> PreprocessorElement:
    return PreprocessorElement(1, of_seq(children))


__all__ = ["ReadHelpers_indentLevel", "write", "read", "mk_line", "mkl_level", "mk_intendation"]




from __future__ import annotations
from typing import Any
from ..fable_library.array_ import (iterate as iterate_1, iterate_indexed)
from ..fable_library.option import some
from ..fable_library.seq import iterate
from ..fable_library.types import Array
from ..fable_library.util import ignore
from ..fable_openpyxl.openpyxl import (Worksheet, Table, Workbook, Cell)
from ..fs_spreadsheet.Cells.fs_cell import FsCell
from ..fs_spreadsheet.fs_worksheet import FsWorksheet
from ..fs_spreadsheet.Tables.fs_table import FsTable
from .cell import (from_fs_cell, to_fs_cell)
from .table import (from_fs_table, to_fs_table)

def from_fs_worksheet(parent: Workbook, fs_ws: FsWorksheet) -> Worksheet:
    py_ws: Worksheet = parent.create_sheet(fs_ws.Name)
    def action(table: FsTable, parent: Any=parent, fs_ws: Any=fs_ws) -> None:
        py_table: Table = from_fs_table(table)
        value: None = py_ws.add_table(py_table)
        ignore(None)

    iterate(action, fs_ws.Tables)
    def action_1(cell: FsCell, parent: Any=parent, fs_ws: Any=fs_ws) -> None:
        py_cell: Any | None = from_fs_cell(cell)
        ignore(py_ws.cell(cell.Address.RowNumber, cell.Address.ColumnNumber, some(py_cell)))

    iterate(action_1, fs_ws.CellCollection.GetCells())
    return py_ws


def to_fs_worksheet(py_ws: Worksheet) -> FsWorksheet:
    fs_ws: FsWorksheet = FsWorksheet(py_ws.title)
    def action(table: Table, py_ws: Any=py_ws) -> None:
        t: FsTable = to_fs_table(table)
        ignore(fs_ws.AddTable(t))

    iterate_1(action, list(py_ws.tables.values()))
    def action_2(row_index: int, row: Array[Cell], py_ws: Any=py_ws) -> None:
        def action_1(col_index: int, cell: Cell, row_index: Any=row_index, row: Any=row) -> None:
            if (type(cell.value).__name__) != "NoneType":
                c: FsCell = to_fs_cell(py_ws.title, row_index + 1, col_index + 1, cell)
                ignore(fs_ws.AddCell(c))


        iterate_indexed(action_1, row)

    iterate_indexed(action_2, [list(inner_tuple) for inner_tuple in py_ws.rows])
    return fs_ws


__all__ = ["from_fs_worksheet", "to_fs_worksheet"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ...arctrl_core.arc_types import ArcRun
from ...arctrl_core.comment import (Comment, Remark)
from ...arctrl_core.conversion import (JsonTypes_decomposeTechnologyPlatform, JsonTypes_composeTechnologyPlatform)
from ...arctrl_core.Helper.collections_ import (Option_fromValueWithDefault, ResizeArray_iter)
from ...arctrl_core.Helper.identifier import (create_missing_identifier, Assay_tryIdentifierFromFileName, Run_fileNameFromIdentifier)
from ...arctrl_core.ontology_annotation import OntologyAnnotation
from ...fable_library.list import (of_array, FSharpList, map as map_1, empty, cons, reverse)
from ...fable_library.map_util import add_to_dict
from ...fable_library.option import (map, default_arg)
from ...fable_library.seq2 import List_distinct
from ...fable_library.string_ import (starts_with_exact, join)
from ...fable_library.types import Array
from ...fable_library.util import (string_hash, IEnumerable_1, IEnumerator)
from .comment import (Comment_fromString, Comment_toString)
from .sparse_table import (SparseTable__TryGetValueDefault_5BAE6133, SparseTable__TryGetValue_11FD62A8, SparseTable, SparseTable_Create_Z2192E64B, SparseTable_FromRows_Z5579EC29, SparseTable_ToRows_759CAFC1)

labels: FSharpList[str] = of_array(["Identifier", "Title", "Description", "Workflow Identifiers", "Measurement Type", "Measurement Type Term Accession Number", "Measurement Type Term Source REF", "Technology Type", "Technology Type Term Accession Number", "Technology Type Term Source REF", "Technology Platform", "File Name"])

def from_string(identifier: str | None, title: str | None, description: str | None, workflow_identifiers: str | None, measurement_type: str | None, measurement_type_term_source_ref: str | None, measurement_type_term_accession_number: str | None, technology_type: str | None, technology_type_term_source_ref: str | None, technology_type_term_accession_number: str | None, technology_platform: str | None, file_name: str | None, comments: Array[Comment]) -> ArcRun:
    workflow_identifiers_1: Array[str]
    if workflow_identifiers is None:
        workflow_identifiers_1 = []

    else: 
        wi: str = workflow_identifiers
        workflow_identifiers_1 = list(wi.split(";"))

    measurement_type_1: OntologyAnnotation = OntologyAnnotation.create(measurement_type, measurement_type_term_source_ref, measurement_type_term_accession_number)
    technology_type_1: OntologyAnnotation = OntologyAnnotation.create(technology_type, technology_type_term_source_ref, technology_type_term_accession_number)
    identifier_4: str
    if identifier is None:
        if file_name is None:
            identifier_4 = create_missing_identifier()

        else: 
            match_value: str | None = Assay_tryIdentifierFromFileName(file_name)
            identifier_4 = match_value if (match_value is not None) else create_missing_identifier()


    else: 
        identifier_4 = identifier

    measurement_type_2: OntologyAnnotation | None = Option_fromValueWithDefault(OntologyAnnotation(), measurement_type_1)
    technology_type_2: OntologyAnnotation | None = Option_fromValueWithDefault(OntologyAnnotation(), technology_type_1)
    def _arrow3486(name: str, identifier: Any=identifier, title: Any=title, description: Any=description, workflow_identifiers: Any=workflow_identifiers, measurement_type: Any=measurement_type, measurement_type_term_source_ref: Any=measurement_type_term_source_ref, measurement_type_term_accession_number: Any=measurement_type_term_accession_number, technology_type: Any=technology_type, technology_type_term_source_ref: Any=technology_type_term_source_ref, technology_type_term_accession_number: Any=technology_type_term_accession_number, technology_platform: Any=technology_platform, file_name: Any=file_name, comments: Any=comments) -> OntologyAnnotation:
        return JsonTypes_decomposeTechnologyPlatform(name)

    technology_platform_1: OntologyAnnotation | None = map(_arrow3486, technology_platform)
    return ArcRun.make(identifier_4, title, description, measurement_type_2, technology_type_2, technology_platform_1, workflow_identifiers_1, [], None, [], None, [], comments)


def from_sparse_table(matrix: SparseTable) -> ArcRun:
    def mapping(k: str, matrix: Any=matrix) -> Comment:
        return Comment_fromString(k, SparseTable__TryGetValueDefault_5BAE6133(matrix, "", (k, 0)))

    comments: Array[Comment] = list(map_1(mapping, matrix.CommentKeys))
    return from_string(SparseTable__TryGetValue_11FD62A8(matrix, ("Identifier", 0)), SparseTable__TryGetValue_11FD62A8(matrix, ("Title", 0)), SparseTable__TryGetValue_11FD62A8(matrix, ("Description", 0)), SparseTable__TryGetValue_11FD62A8(matrix, ("Workflow Identifiers", 0)), SparseTable__TryGetValue_11FD62A8(matrix, ("Measurement Type", 0)), SparseTable__TryGetValue_11FD62A8(matrix, ("Measurement Type Term Source REF", 0)), SparseTable__TryGetValue_11FD62A8(matrix, ("Measurement Type Term Accession Number", 0)), SparseTable__TryGetValue_11FD62A8(matrix, ("Technology Type", 0)), SparseTable__TryGetValue_11FD62A8(matrix, ("Technology Type Term Source REF", 0)), SparseTable__TryGetValue_11FD62A8(matrix, ("Technology Type Term Accession Number", 0)), SparseTable__TryGetValue_11FD62A8(matrix, ("Technology Platform", 0)), SparseTable__TryGetValue_11FD62A8(matrix, ("File Name", 0)), comments)


def to_sparse_table(run: ArcRun) -> SparseTable:
    matrix: SparseTable = SparseTable_Create_Z2192E64B(None, labels, None, 2)
    comment_keys: FSharpList[str] = empty()
    pattern_input: tuple[str, str] = (("", "")) if starts_with_exact(run.Identifier, "MISSING_IDENTIFIER_") else ((run.Identifier, Run_fileNameFromIdentifier(run.Identifier)))
    workflow_identifiers: str = join(";", run.WorkflowIdentifiers)
    mt_1: dict[str, Any]
    mt: OntologyAnnotation = default_arg(run.MeasurementType, OntologyAnnotation())
    mt_1 = OntologyAnnotation.to_string_object(mt, True)
    tt_1: dict[str, Any]
    tt: OntologyAnnotation = default_arg(run.TechnologyType, OntologyAnnotation())
    tt_1 = OntologyAnnotation.to_string_object(tt, True)
    add_to_dict(matrix.Matrix, ("Identifier", 1), pattern_input[0])
    add_to_dict(matrix.Matrix, ("Title", 1), default_arg(run.Title, ""))
    add_to_dict(matrix.Matrix, ("Description", 1), default_arg(run.Description, ""))
    add_to_dict(matrix.Matrix, ("Workflow Identifiers", 1), workflow_identifiers)
    add_to_dict(matrix.Matrix, ("Measurement Type", 1), mt_1["TermName"])
    add_to_dict(matrix.Matrix, ("Measurement Type Term Accession Number", 1), mt_1["TermAccessionNumber"])
    add_to_dict(matrix.Matrix, ("Measurement Type Term Source REF", 1), mt_1["TermSourceREF"])
    add_to_dict(matrix.Matrix, ("Technology Type", 1), tt_1["TermName"])
    add_to_dict(matrix.Matrix, ("Technology Type Term Accession Number", 1), tt_1["TermAccessionNumber"])
    add_to_dict(matrix.Matrix, ("Technology Type Term Source REF", 1), tt_1["TermSourceREF"])
    def mapping(tp: OntologyAnnotation, run: Any=run) -> str:
        return JsonTypes_composeTechnologyPlatform(tp)

    add_to_dict(matrix.Matrix, ("Technology Platform", 1), default_arg(map(mapping, run.TechnologyPlatform), ""))
    add_to_dict(matrix.Matrix, ("File Name", 1), pattern_input[1])
    def f(comment: Comment, run: Any=run) -> None:
        nonlocal comment_keys
        pattern_input_1: tuple[str, str] = Comment_toString(comment)
        n: str = pattern_input_1[0]
        comment_keys = cons(n, comment_keys)
        add_to_dict(matrix.Matrix, (n, 1), pattern_input_1[1])

    ResizeArray_iter(f, run.Comments)
    class ObjectExpr3488:
        @property
        def Equals(self) -> Callable[[str, str], bool]:
            def _arrow3487(x: str, y: str) -> bool:
                return x == y

            return _arrow3487

        @property
        def GetHashCode(self) -> Callable[[str], int]:
            return string_hash

    return SparseTable(matrix.Matrix, matrix.Keys, reverse(List_distinct(comment_keys, ObjectExpr3488())), matrix.ColumnCount)


def from_rows(line_number: int, rows: IEnumerator[IEnumerable_1[tuple[int, str]]]) -> tuple[str | None, int, FSharpList[Remark], ArcRun]:
    tupled_arg: tuple[str | None, int, FSharpList[Remark], SparseTable] = SparseTable_FromRows_Z5579EC29(rows, labels, line_number, "Run")
    return (tupled_arg[0], tupled_arg[1], tupled_arg[2], from_sparse_table(tupled_arg[3]))


def to_rows(run: ArcRun) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
    return SparseTable_ToRows_759CAFC1(to_sparse_table(run), "Run")


__all__ = ["labels", "from_string", "from_sparse_table", "to_sparse_table", "from_rows", "to_rows"]




from __future__ import annotations
from typing import (Any, TypeVar)
from ...fable_library.option import default_arg
from ...fable_library.seq import map
from ...fable_library.util import (to_enumerable, IEnumerable_1)
from ...thoth_json_core.decode import (object, IOptionalGetter, int_1, IGetters)
from ...thoth_json_core.types import (IEncodable, IEncoderHelpers_1, Decoder_1)
from ..Cells.fs_cell import (FsCell, DataType)
from ..fs_address import FsAddress
from .value import (encode, decode)

__A_ = TypeVar("__A_")

def encode_no_number(cell: FsCell) -> IEncodable:
    values: IEnumerable_1[tuple[str, IEncodable]] = to_enumerable([("value", encode(cell.Value))])
    class ObjectExpr329(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any], cell: Any=cell) -> Any:
            def mapping(tupled_arg: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg[0], tupled_arg[1].Encode(helpers))

            arg: IEnumerable_1[tuple[str, __A_]] = map(mapping, values)
            return helpers.encode_object(arg)

    return ObjectExpr329()


def encode_rows(cell: FsCell) -> IEncodable:
    def _arrow331(__unit: None=None, cell: Any=cell) -> IEncodable:
        value_1: int = cell.ColumnNumber or 0
        class ObjectExpr330(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                return helpers.encode_signed_integral_number(value_1)

        return ObjectExpr330()

    values: IEnumerable_1[tuple[str, IEncodable]] = to_enumerable([("column", _arrow331()), ("value", encode(cell.Value))])
    class ObjectExpr332(IEncodable):
        def Encode(self, helpers_1: IEncoderHelpers_1[Any], cell: Any=cell) -> Any:
            def mapping(tupled_arg: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg[0], tupled_arg[1].Encode(helpers_1))

            arg: IEnumerable_1[tuple[str, __A_]] = map(mapping, values)
            return helpers_1.encode_object(arg)

    return ObjectExpr332()


def decode_rows(row_number: int | None=None) -> Decoder_1[FsCell]:
    def _arrow335(builder: IGetters, row_number: Any=row_number) -> FsCell:
        def _arrow333(__unit: None=None) -> tuple[Any, DataType] | None:
            object_arg: IOptionalGetter = builder.Optional
            return object_arg.Field("value", decode)

        pattern_input: tuple[Any, DataType] = default_arg(_arrow333(), ("", DataType(4)))
        def _arrow334(__unit: None=None) -> int | None:
            object_arg_1: IOptionalGetter = builder.Optional
            return object_arg_1.Field("column", int_1)

        c: int = default_arg(_arrow334(), 0) or 0
        return FsCell(pattern_input[0], pattern_input[1], FsAddress(default_arg(row_number, 0), c))

    return object(_arrow335)


def encode_cols(cell: FsCell) -> IEncodable:
    def _arrow337(__unit: None=None, cell: Any=cell) -> IEncodable:
        value_1: int = cell.RowNumber or 0
        class ObjectExpr336(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                return helpers.encode_signed_integral_number(value_1)

        return ObjectExpr336()

    values: IEnumerable_1[tuple[str, IEncodable]] = to_enumerable([("row", _arrow337()), ("value", encode(cell.Value))])
    class ObjectExpr338(IEncodable):
        def Encode(self, helpers_1: IEncoderHelpers_1[Any], cell: Any=cell) -> Any:
            def mapping(tupled_arg: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg[0], tupled_arg[1].Encode(helpers_1))

            arg: IEnumerable_1[tuple[str, __A_]] = map(mapping, values)
            return helpers_1.encode_object(arg)

    return ObjectExpr338()


def decode_cols(col_number: int | None=None) -> Decoder_1[FsCell]:
    def _arrow341(builder: IGetters, col_number: Any=col_number) -> FsCell:
        def _arrow339(__unit: None=None) -> tuple[Any, DataType] | None:
            object_arg: IOptionalGetter = builder.Optional
            return object_arg.Field("value", decode)

        pattern_input: tuple[Any, DataType] = default_arg(_arrow339(), ("", DataType(4)))
        def _arrow340(__unit: None=None) -> int | None:
            object_arg_1: IOptionalGetter = builder.Optional
            return object_arg_1.Field("row", int_1)

        return FsCell(pattern_input[0], pattern_input[1], FsAddress(default_arg(_arrow340(), 0), default_arg(col_number, 0)))

    return object(_arrow341)


__all__ = ["encode_no_number", "encode_rows", "decode_rows", "encode_cols", "decode_cols"]




from __future__ import annotations

from collections.abc import Callable
from typing import Any, Generic, TypeVar

from .map_util import get_item_from_dict, try_get_value
from .option import some
from .reflection import TypeInfo, class_type
from .resize_array import find_index
from .seq import concat, iterate, iterate_indexed, map
from .types import Array, FSharpRef
from .util import IEnumerable_1, IEnumerator, IEqualityComparer_1, dispose, get_enumerator, ignore, to_iterator


_T = TypeVar("_T")


def _expr22(gen0: TypeInfo) -> TypeInfo:
    return class_type("Fable.Collections.HashSet", [gen0], HashSet)


class HashSet(Generic[_T]):
    def __init__(self, items: IEnumerable_1[Any], comparer: IEqualityComparer_1[Any]) -> None:
        this: FSharpRef[HashSet[_T]] = FSharpRef(None)
        self.comparer: IEqualityComparer_1[Any] = comparer
        this.contents = self
        self.hash_map: Any = dict([])
        self.init_00409: int = 1
        with get_enumerator(items) as enumerator:
            while enumerator.System_Collections_IEnumerator_MoveNext():
                item: _T = enumerator.System_Collections_Generic_IEnumerator_1_get_Current()
                ignore(HashSet__Add_2B595(this.contents, item))

    @property
    def Symbol_toStringTag(self, __unit: None = None) -> str:
        return "HashSet"

    def to_json(self, __unit: None = None) -> Any:
        this: HashSet[_T] = self
        return list(this)

    def System_Collections_IEnumerable_GetEnumerator(self, __unit: None = None) -> IEnumerator[Any]:
        this: HashSet[_T] = self
        return get_enumerator(this)

    def GetEnumerator(self, __unit: None = None) -> IEnumerator[_T]:
        this: HashSet[_T] = self
        return get_enumerator(concat(this.hash_map.values()))

    def __iter__(self) -> IEnumerator[_T]:
        return to_iterator(self.GetEnumerator())

    def System_Collections_Generic_ICollection_1_Add2B595(self, item: _T | None = None) -> None:
        this: HashSet[_T] = self
        ignore(HashSet__Add_2B595(this, item))

    def System_Collections_Generic_ICollection_1_Clear(self, __unit: None = None) -> None:
        this: HashSet[_T] = self
        HashSet__Clear(this)

    def System_Collections_Generic_ICollection_1_Contains2B595(self, item: _T | None = None) -> bool:
        this: HashSet[_T] = self
        return HashSet__Contains_2B595(this, item)

    def System_Collections_Generic_ICollection_1_CopyToZ3B4C077E(self, array: Array[_T], array_index: int) -> None:
        this: HashSet[_T] = self

        def action(i: int, e: _T) -> None:
            array[array_index + i] = e

        iterate_indexed(action, this)

    def __len__(self, __unit: None = None) -> int:
        this: HashSet[_T] = self
        return HashSet__get_Count(this)

    def System_Collections_Generic_ICollection_1_get_IsReadOnly(self, __unit: None = None) -> bool:
        return False

    def System_Collections_Generic_ICollection_1_Remove2B595(self, item: _T | None = None) -> bool:
        this: HashSet[_T] = self
        return HashSet__Remove_2B595(this, item)

    @property
    def size(self, __unit: None = None) -> int:
        this: HashSet[_T] = self
        return HashSet__get_Count(this)

    def add(self, k: _T | None = None) -> Set_1[_T]:
        this: HashSet[_T] = self
        ignore(HashSet__Add_2B595(this, k))
        return this

    def clear(self, __unit: None = None) -> None:
        this: HashSet[_T] = self
        HashSet__Clear(this)

    def __delitem__(self, k: _T | None = None) -> bool:
        this: HashSet[_T] = self
        return HashSet__Remove_2B595(this, k)

    def __contains__(self, k: _T | None = None) -> bool:
        this: HashSet[_T] = self
        return HashSet__Contains_2B595(this, k)

    def keys(self, __unit: None = None) -> IEnumerable_1[_T]:
        this: HashSet[_T] = self

        def mapping(x: _T | None = None) -> _T | None:
            return x

        return map(mapping, this)

    def values(self, __unit: None = None) -> IEnumerable_1[_T]:
        this: HashSet[_T] = self

        def mapping(x: _T | None = None) -> _T | None:
            return x

        return map(mapping, this)

    def entries(self, __unit: None = None) -> IEnumerable_1[tuple[_T, _T]]:
        this: HashSet[_T] = self

        def mapping(v: _T | None = None) -> tuple[_T, _T]:
            return (v, v)

        return map(mapping, this)

    def for_each(self, f: Callable[[_T, _T, Set_1[_T]], None], this_arg: Any | None = None) -> None:
        this: HashSet[_T] = self

        def action(x: _T | None = None) -> None:
            f(x, x, this)

        iterate(action, this)


HashSet_reflection = _expr22


def HashSet__ctor_Z6150332D(items: IEnumerable_1[Any], comparer: IEqualityComparer_1[Any]) -> HashSet[_T]:
    return HashSet(items, comparer)


def HashSet__TryFindIndex_2B595(this: HashSet[Any], k: Any) -> tuple[bool, int, int]:
    h: int = this.comparer.GetHashCode(k) or 0
    match_value: tuple[bool, Array[_T]]
    out_arg: Array[_T] = None

    def _arrow23(__unit: None = None, this: Any = this, k: Any = k) -> Array[_T]:
        return out_arg

    def _arrow24(v: Array[_T], this: Any = this, k: Any = k) -> None:
        nonlocal out_arg
        out_arg = v

    match_value = (try_get_value(this.hash_map, h, FSharpRef(_arrow23, _arrow24)), out_arg)
    if match_value[0]:

        def _arrow25(v_1: _T | None = None, this: Any = this, k: Any = k) -> bool:
            return this.comparer.Equals(k, v_1)

        return (True, h, find_index(_arrow25, match_value[1]))

    else:
        return (False, h, -1)


def HashSet__TryFind_2B595(this: HashSet[Any], k: Any) -> Any | None:
    match_value: tuple[bool, int, int] = HashSet__TryFindIndex_2B595(this, k)
    (pattern_matching_result,) = (None,)
    if match_value[0]:
        if match_value[2] > -1:
            pattern_matching_result = 0

        else:
            pattern_matching_result = 1

    else:
        pattern_matching_result = 1

    if pattern_matching_result == 0:
        return some(get_item_from_dict(this.hash_map, match_value[1])[match_value[2]])

    elif pattern_matching_result == 1:
        return None


def HashSet__get_Comparer(this: HashSet[Any]) -> IEqualityComparer_1[Any]:
    return this.comparer


def HashSet__Clear(this: HashSet[Any]) -> None:
    this.hash_map.clear()


def HashSet__get_Count(this: HashSet[Any]) -> int:
    count: int = 0
    enumerator: Any = get_enumerator(this.hash_map.values())
    try:
        while enumerator.System_Collections_IEnumerator_MoveNext():
            items: Array[_T] = enumerator.System_Collections_Generic_IEnumerator_1_get_Current()
            count = (count + len(items)) or 0

    finally:
        dispose(enumerator)

    return count


def HashSet__Add_2B595(this: HashSet[Any], k: Any) -> bool:
    match_value: tuple[bool, int, int] = HashSet__TryFindIndex_2B595(this, k)
    if match_value[0]:
        if match_value[2] > -1:
            return False

        else:
            value: None = get_item_from_dict(this.hash_map, match_value[1]).append(k)
            ignore(None)
            return True

    else:
        this.hash_map[match_value[1]] = [k]
        return True


def HashSet__Contains_2B595(this: HashSet[Any], k: Any) -> bool:
    match_value: tuple[bool, int, int] = HashSet__TryFindIndex_2B595(this, k)
    (pattern_matching_result,) = (None,)
    if match_value[0]:
        if match_value[2] > -1:
            pattern_matching_result = 0

        else:
            pattern_matching_result = 1

    else:
        pattern_matching_result = 1

    if pattern_matching_result == 0:
        return True

    elif pattern_matching_result == 1:
        return False


def HashSet__Remove_2B595(this: HashSet[Any], k: Any) -> bool:
    match_value: tuple[bool, int, int] = HashSet__TryFindIndex_2B595(this, k)
    (pattern_matching_result,) = (None,)
    if match_value[0]:
        if match_value[2] > -1:
            pattern_matching_result = 0

        else:
            pattern_matching_result = 1

    else:
        pattern_matching_result = 1

    if pattern_matching_result == 0:
        get_item_from_dict(this.hash_map, match_value[1]).pop(match_value[2])
        return True

    elif pattern_matching_result == 1:
        return False


__all__ = [
    "HashSet_reflection",
    "HashSet__TryFindIndex_2B595",
    "HashSet__TryFind_2B595",
    "HashSet__get_Comparer",
    "HashSet__Clear",
    "HashSet__get_Count",
    "HashSet__Add_2B595",
    "HashSet__Contains_2B595",
    "HashSet__Remove_2B595",
]



from abc import abstractmethod
from typing import (Any, Protocol)

class IDisplay(Protocol):
    @abstractmethod
    def Markdown(self, data: str) -> None:
        ...

    @abstractmethod
    def display(self, value: Any) -> None:
        ...





from typing import Any
from ..fable_library.map_util import add_to_dict
from ..fable_library.reg_exp import (replace as replace_1, get_item, groups, create)
from ..fable_library.string_ import (replace, to_text, printf, split)
from ..fable_library.types import Array

def encoding_clean_up(s: str) -> str:
    return replace(s, "\r\n", "\n")


def string_clean_up(dict_1: Any, s: str) -> str:
    n: int = 0
    def matcheval(m: Any, dict_1: Any=dict_1, s: Any=s) -> str:
        nonlocal n
        if get_item(groups(m), "iscomment") is not None:
            return get_item(groups(m), "all") or ""

        else: 
            v: str = get_item(groups(m), "stringValue") or ""
            current_n: int = n or 0
            n = (n + 1) or 0
            add_to_dict(dict_1, current_n, v)
            return to_text(printf("<s f=%i/>"))(current_n)


    return replace_1(create("(?P<all>(?P<iscomment>#.*?)?\"(?P<stringValue>.+)\\\")"), s, matcheval)


def comment_clean_up(dict_1: Any, s: str) -> str:
    n: int = 0
    def matcheval(m: Any, dict_1: Any=dict_1, s: Any=s) -> str:
        nonlocal n
        v: str = get_item(groups(m), "comment") or ""
        current_n: int = n or 0
        n = (n + 1) or 0
        add_to_dict(dict_1, current_n, v)
        return to_text(printf("<c f=%i/>"))(current_n)

    return replace_1(create("#(?P<comment>.*)"), s, matcheval)


def cut(yaml_string: str) -> Array[str]:
    return split(yaml_string, ["\n"], None, 1)


def pipeline(yaml_string: str) -> dict[str, Any]:
    string_map: Any = dict([])
    comment_map: Any = dict([])
    return {
        "CommentMap": comment_map,
        "Lines": cut(comment_clean_up(comment_map, string_clean_up(string_map, encoding_clean_up(yaml_string)))),
        "StringMap": string_map
    }


__all__ = ["encoding_clean_up", "string_clean_up", "comment_clean_up", "cut", "pipeline"]




from __future__ import annotations
from collections.abc import Callable
from dataclasses import dataclass
from typing import Any
from ...fable_library.option import (map, default_arg)
from ...fable_library.reflection import (TypeInfo, option_type, record_type)
from ...fable_library.reg_exp import (get_item, groups)
from ...fable_library.string_ import (to_text, printf)
from ...fable_library.types import (Record, Array)
from ..comment import Comment
from ..Helper.collections_ import Option_fromValueWithDefault
from ..Helper.regex import ActivePatterns__007CRegex_007C__007C
from ..ontology_annotation import (OntologyAnnotation, OntologyAnnotation_reflection)
from ..value import (Value, Value_reflection)

def _expr762() -> TypeInfo:
    return record_type("ARCtrl.Process.Component", [], Component, lambda: [("ComponentValue", option_type(Value_reflection())), ("ComponentUnit", option_type(OntologyAnnotation_reflection())), ("ComponentType", option_type(OntologyAnnotation_reflection()))])


@dataclass(eq = False, repr = False, slots = True)
class Component(Record):
    ComponentValue: Value | None
    ComponentUnit: OntologyAnnotation | None
    ComponentType: OntologyAnnotation | None
    def AlternateName(self, __unit: None=None) -> str | None:
        this: Component = self
        return Component__get_ComponentName(this)

    def MeasurementMethod(self, __unit: None=None) -> str | None:
        return None

    def Description(self, __unit: None=None) -> str | None:
        return None

    def GetCategory(self, __unit: None=None) -> OntologyAnnotation | None:
        this: Component = self
        return this.ComponentType

    def GetAdditionalType(self, __unit: None=None) -> str:
        return "Component"

    def GetValue(self, __unit: None=None) -> Value | None:
        this: Component = self
        return this.ComponentValue

    def GetUnit(self, __unit: None=None) -> OntologyAnnotation | None:
        this: Component = self
        return this.ComponentUnit


Component_reflection = _expr762

def Component__get_ComponentName(this: Component) -> str | None:
    def mapping(v: Value, this: Any=this) -> str:
        return Component_composeName(v, this.ComponentUnit)

    return map(mapping, this.ComponentValue)


def Component_make(value: Value | None=None, unit: OntologyAnnotation | None=None, component_type: OntologyAnnotation | None=None) -> Component:
    return Component(value, unit, component_type)


def Component_create_Z2F0B38C7(value: Value | None=None, unit: OntologyAnnotation | None=None, component_type: OntologyAnnotation | None=None) -> Component:
    return Component_make(value, unit, component_type)


def Component_get_empty(__unit: None=None) -> Component:
    return Component_create_Z2F0B38C7()


def Component_composeName(value: Value, unit: OntologyAnnotation | None=None) -> str:
    if value.tag == 0:
        oa: OntologyAnnotation = value.fields[0]
        return ((("" + oa.NameText) + " (") + oa.TermAccessionShort) + ")"

    elif unit is not None:
        u: OntologyAnnotation = unit
        return ((((("" + value.Text) + " ") + u.NameText) + " (") + u.TermAccessionShort) + ")"

    else: 
        return ("" + value.Text) + ""



def Component_decomposeName_Z721C83C5(name: str) -> tuple[Value, OntologyAnnotation | None]:
    active_pattern_result: Any | None = ActivePatterns__007CRegex_007C__007C("^(?<value>[\\d\\.]+) (?<unit>.+) \\((?<ontology>[^(]*:[^)]*)\\)", name)
    if active_pattern_result is not None:
        unitr: Any = active_pattern_result
        oa: OntologyAnnotation
        tan: str = get_item(groups(unitr), "ontology") or ""
        oa = OntologyAnnotation.from_term_annotation(tan)
        v: Value
        value: str = get_item(groups(unitr), "value") or ""
        v = Value.from_string(value)
        u: str = get_item(groups(unitr), "unit") or ""
        oa.Name = u
        return (v, oa)

    else: 
        active_pattern_result_1: Any | None = ActivePatterns__007CRegex_007C__007C("^(?<value>[^\\(]+) \\((?<ontology>[^(]*:[^)]*)\\)", name)
        if active_pattern_result_1 is not None:
            r: Any = active_pattern_result_1
            oa_1: OntologyAnnotation
            tan_1: str = get_item(groups(r), "ontology") or ""
            oa_1 = OntologyAnnotation.from_term_annotation(tan_1)
            v_1: str = get_item(groups(r), "value") or ""
            oa_1.Name = v_1
            return (Value(0, oa_1), None)

        else: 
            active_pattern_result_2: Any | None = ActivePatterns__007CRegex_007C__007C("^(?<value>[^\\(\\)]+) \\(\\)", name)
            if active_pattern_result_2 is not None:
                r_1: Any = active_pattern_result_2
                return (Value(0, OntologyAnnotation(get_item(groups(r_1), "value") or "")), None)

            else: 
                return (Value(3, name), None)





def Component_fromISAString_7C9A7CF8(name: str | None=None, term: str | None=None, source: str | None=None, accession: str | None=None, comments: Array[Comment] | None=None) -> Component:
    c_type: OntologyAnnotation | None
    v: OntologyAnnotation = OntologyAnnotation.create(term, source, accession, comments)
    c_type = Option_fromValueWithDefault(OntologyAnnotation(), v)
    if name is None:
        return Component_make(None, None, c_type)

    else: 
        pattern_input: tuple[Value, OntologyAnnotation | None] = Component_decomposeName_Z721C83C5(name)
        return Component_make(Option_fromValueWithDefault(Value(3, ""), pattern_input[0]), pattern_input[1], c_type)



def Component_toStringObject_Z685B8F25(c: Component) -> tuple[str, dict[str, Any]]:
    oa_1: dict[str, Any]
    value: dict[str, Any] = {
        "TermAccessionNumber": "",
        "TermName": "",
        "TermSourceREF": ""
    }
    def mapping(oa: OntologyAnnotation, c: Any=c) -> dict[str, Any]:
        return OntologyAnnotation.to_string_object(oa)

    oa_1 = default_arg(map(mapping, c.ComponentType), value)
    return (default_arg(Component__get_ComponentName(c), ""), oa_1)


def Component__get_NameText(this: Component) -> str:
    def mapping(c: OntologyAnnotation, this: Any=this) -> str:
        return c.NameText

    return default_arg(map(mapping, this.ComponentType), "")


def Component__get_UnitText(this: Component) -> str:
    def mapping(c: OntologyAnnotation, this: Any=this) -> str:
        return c.NameText

    return default_arg(map(mapping, this.ComponentUnit), "")


def Component__get_ValueText(this: Component) -> str:
    def mapping(c: Value, this: Any=this) -> str:
        return c.Text

    return default_arg(map(mapping, this.ComponentValue), "")


def Component__get_ValueWithUnitText(this: Component) -> str:
    def mapping(oa: OntologyAnnotation, this: Any=this) -> str:
        return oa.NameText

    unit: str | None = map(mapping, this.ComponentUnit)
    v: str = Component__get_ValueText(this)
    if unit is None:
        return v

    else: 
        u: str = unit
        return to_text(printf("%s %s"))(v)(u)



def Component__MapCategory_658CFBF6(this: Component, f: Callable[[OntologyAnnotation], OntologyAnnotation]) -> Component:
    return Component(this.ComponentValue, this.ComponentUnit, map(f, this.ComponentType))


def Component__SetCategory_ZDED3A0F(this: Component, c: OntologyAnnotation) -> Component:
    return Component(this.ComponentValue, this.ComponentUnit, c)


def Component_createAsPV(alternate_name: str | None=None, measurement_method: str | None=None, description: str | None=None, category: OntologyAnnotation | None=None, value: Value | None=None, unit: OntologyAnnotation | None=None) -> Component:
    return Component_create_Z2F0B38C7(value, unit, category)


__all__ = ["Component_reflection", "Component__get_ComponentName", "Component_make", "Component_create_Z2F0B38C7", "Component_get_empty", "Component_composeName", "Component_decomposeName_Z721C83C5", "Component_fromISAString_7C9A7CF8", "Component_toStringObject_Z685B8F25", "Component__get_NameText", "Component__get_UnitText", "Component__get_ValueText", "Component__get_ValueWithUnitText", "Component__MapCategory_658CFBF6", "Component__SetCategory_ZDED3A0F", "Component_createAsPV"]




from __future__ import annotations
from typing import Any
from ..arctrl_core.Helper.hash_codes import (box_hash_array, box_hash_option)
from ..fable_library.option import value as value_1
from ..fable_library.reflection import (TypeInfo, class_type)
from ..fable_library.seq import (to_list, delay, append, singleton, empty)
from ..fable_library.string_ import join
from ..fable_library.util import (IEnumerable_1, equals, string_hash)

def _expr994() -> TypeInfo:
    return class_type("ARCtrl.ValidationPackages.ValidationPackage", None, ValidationPackage)


class ValidationPackage:
    def __init__(self, name: str, version: str | None=None) -> None:
        self.version: str | None = version
        self._name: str = name
        self._version: str | None = self.version

    @property
    def Name(self, __unit: None=None) -> str:
        this: ValidationPackage = self
        return this._name

    @Name.setter
    def Name(self, name: str) -> None:
        this: ValidationPackage = self
        this._name = name

    @property
    def Version(self, __unit: None=None) -> str | None:
        this: ValidationPackage = self
        return this._version

    @Version.setter
    def Version(self, version: str | None=None) -> None:
        this: ValidationPackage = self
        this._version = version

    @staticmethod
    def make(name: str, version: str | None=None) -> ValidationPackage:
        return ValidationPackage(name, version)

    def Copy(self, __unit: None=None) -> ValidationPackage:
        this: ValidationPackage = self
        name: str = this.Name
        version: str | None = this.Version
        return ValidationPackage.make(name, version)

    def __str__(self, __unit: None=None) -> str:
        this: ValidationPackage = self
        def _arrow993(__unit: None=None) -> IEnumerable_1[str]:
            def _arrow992(__unit: None=None) -> IEnumerable_1[str]:
                def _arrow991(__unit: None=None) -> IEnumerable_1[str]:
                    def _arrow990(__unit: None=None) -> IEnumerable_1[str]:
                        return singleton("}")

                    return append(singleton((" Version = " + value_1(this.Version)) + "") if (this.version is not None) else empty(), delay(_arrow990))

                return append(singleton((" Name = " + this.Name) + ""), delay(_arrow991))

            return append(singleton("{"), delay(_arrow992))

        return join("\n", to_list(delay(_arrow993)))

    def __eq__(self, obj: Any=None) -> bool:
        this: ValidationPackage = self
        return (equals(obj.Version, this.Version) if (obj.Name == this.Name) else False) if isinstance(obj, ValidationPackage) else False

    def __hash__(self, __unit: None=None) -> Any:
        this: ValidationPackage = self
        return box_hash_array([string_hash(this.Name), box_hash_option(this.Version)])


ValidationPackage_reflection = _expr994

def ValidationPackage__ctor_27AED5E3(name: str, version: str | None=None) -> ValidationPackage:
    return ValidationPackage(name, version)


__all__ = ["ValidationPackage_reflection"]




from __future__ import annotations
from dataclasses import dataclass
from typing import Any
from ..dynamic_obj.dynamic_obj import (DynamicObj, DynamicObj_reflection)
from ..dynamic_obj.dyn_obj import set_property
from ..fable_library.reflection import (TypeInfo, class_type, string_type, option_type, bool_type, record_type, union_type, array_type)
from ..fable_library.types import (Record, Array, Union, FSharpRef)
from ..fable_library.util import identity_hash

def _expr473() -> TypeInfo:
    return class_type("ARCtrl.CWL.FileInstance", None, FileInstance, DynamicObj_reflection())


class FileInstance(DynamicObj):
    def __init__(self, __unit: None=None) -> None:
        super().__init__()
        pass

    def __hash__(self, __unit: None=None) -> int:
        this: FileInstance = self
        return identity_hash(this.DeepCopyProperties())

    def __eq__(self, o: Any=None) -> bool:
        this: FileInstance = self
        return this.StructurallyEquals(o) if isinstance(o, FileInstance) else False


FileInstance_reflection = _expr473

def FileInstance__ctor(__unit: None=None) -> FileInstance:
    return FileInstance(__unit)


def _expr474() -> TypeInfo:
    return class_type("ARCtrl.CWL.DirectoryInstance", None, DirectoryInstance, DynamicObj_reflection())


class DirectoryInstance(DynamicObj):
    def __init__(self, __unit: None=None) -> None:
        super().__init__()
        pass

    def __eq__(self, o: Any=None) -> bool:
        this: DirectoryInstance = self
        return this.StructurallyEquals(o) if isinstance(o, DirectoryInstance) else False

    def __hash__(self, __unit: None=None) -> int:
        this: DirectoryInstance = self
        return identity_hash(this.DeepCopyProperties())


DirectoryInstance_reflection = _expr474

def DirectoryInstance__ctor(__unit: None=None) -> DirectoryInstance:
    return DirectoryInstance(__unit)


def _expr475() -> TypeInfo:
    return record_type("ARCtrl.CWL.DirentInstance", [], DirentInstance, lambda: [("Entry", string_type), ("Entryname", option_type(string_type)), ("Writable", option_type(bool_type))])


@dataclass(eq = False, repr = False, slots = True)
class DirentInstance(Record):
    Entry: str
    Entryname: str | None
    Writable: bool | None

DirentInstance_reflection = _expr475

def _expr476() -> TypeInfo:
    return union_type("ARCtrl.CWL.CWLType", [], CWLType, lambda: [[("Item", FileInstance_reflection())], [("Item", DirectoryInstance_reflection())], [("Item", DirentInstance_reflection())], [], [], [], [], [], [], [], [], [("Item", CWLType_reflection())]])


class CWLType(Union):
    def __init__(self, tag: int, *fields: Any) -> None:
        super().__init__()
        self.tag: int = tag or 0
        self.fields: Array[Any] = list(fields)

    @staticmethod
    def cases() -> list[str]:
        return ["File", "Directory", "Dirent", "String", "Int", "Long", "Float", "Double", "Boolean", "Stdout", "Null", "Array"]


CWLType_reflection = _expr476

def CWLType_file(__unit: None=None) -> CWLType:
    return CWLType(0, FileInstance__ctor())


def CWLType_directory(__unit: None=None) -> CWLType:
    return CWLType(1, DirectoryInstance__ctor())


def _expr477() -> TypeInfo:
    return class_type("ARCtrl.CWL.InputRecordSchema", None, InputRecordSchema, DynamicObj_reflection())


class InputRecordSchema(DynamicObj):
    def __init__(self, __unit: None=None) -> None:
        super().__init__()
        pass


InputRecordSchema_reflection = _expr477

def InputRecordSchema__ctor(__unit: None=None) -> InputRecordSchema:
    return InputRecordSchema(__unit)


def _expr478() -> TypeInfo:
    return class_type("ARCtrl.CWL.InputEnumSchema", None, InputEnumSchema, DynamicObj_reflection())


class InputEnumSchema(DynamicObj):
    def __init__(self, __unit: None=None) -> None:
        super().__init__()
        pass


InputEnumSchema_reflection = _expr478

def InputEnumSchema__ctor(__unit: None=None) -> InputEnumSchema:
    return InputEnumSchema(__unit)


def _expr479() -> TypeInfo:
    return class_type("ARCtrl.CWL.InputArraySchema", None, InputArraySchema, DynamicObj_reflection())


class InputArraySchema(DynamicObj):
    def __init__(self, __unit: None=None) -> None:
        super().__init__()
        pass


InputArraySchema_reflection = _expr479

def InputArraySchema__ctor(__unit: None=None) -> InputArraySchema:
    return InputArraySchema(__unit)


def _expr480() -> TypeInfo:
    return class_type("ARCtrl.CWL.SchemaDefRequirementType", None, SchemaDefRequirementType, DynamicObj_reflection())


class SchemaDefRequirementType(DynamicObj):
    def __init__(self, types: Any=None, definitions: Any=None) -> None:
        super().__init__()
        this: FSharpRef[SchemaDefRequirementType] = FSharpRef(None)
        this.contents = self
        self.init_004069: int = 1
        set_property("types", definitions, this.contents)


SchemaDefRequirementType_reflection = _expr480

def SchemaDefRequirementType__ctor_541DA560(types: Any=None, definitions: Any=None) -> SchemaDefRequirementType:
    return SchemaDefRequirementType(types, definitions)


def _expr481() -> TypeInfo:
    return record_type("ARCtrl.CWL.SoftwarePackage", [], SoftwarePackage, lambda: [("Package", string_type), ("Version", option_type(array_type(string_type))), ("Specs", option_type(array_type(string_type)))])


@dataclass(eq = False, repr = False, slots = True)
class SoftwarePackage(Record):
    Package: str
    Version: Array[str] | None
    Specs: Array[str] | None

SoftwarePackage_reflection = _expr481

__all__ = ["FileInstance_reflection", "DirectoryInstance_reflection", "DirentInstance_reflection", "CWLType_reflection", "CWLType_file", "CWLType_directory", "InputRecordSchema_reflection", "InputEnumSchema_reflection", "InputArraySchema_reflection", "SchemaDefRequirementType_reflection", "SoftwarePackage_reflection"]




from __future__ import annotations
from typing import (Any, TypeVar)
from ...arctrl_core.data_context import DataContext
from ...arctrl_core.data_map import DataMap
from ...arctrl_core.ontology_annotation import OntologyAnnotation
from ...arctrl_core.Table.composite_cell import CompositeCell
from ...fable_library.seq import map
from ...fable_library.types import Array
from ...fable_library.util import (to_enumerable, IEnumerable_1)
from ...thoth_json_core.decode import (object, resize_array, IRequiredGetter, IGetters)
from ...thoth_json_core.encode import seq
from ...thoth_json_core.types import (IEncodable, IEncoderHelpers_1, Decoder_1)
from .data_context import (encoder as encoder_1, decoder as decoder_1)

__A_ = TypeVar("__A_")

def encoder(dm: DataMap) -> IEncodable:
    def mapping(dc: DataContext, dm: Any=dm) -> IEncodable:
        return encoder_1(dc)

    values_1: IEnumerable_1[tuple[str, IEncodable]] = to_enumerable([("dataContexts", seq(map(mapping, dm.DataContexts)))])
    class ObjectExpr2875(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any], dm: Any=dm) -> Any:
            def mapping_1(tupled_arg: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg[0], tupled_arg[1].Encode(helpers))

            arg: IEnumerable_1[tuple[str, __A_]] = map(mapping_1, values_1)
            return helpers.encode_object(arg)

    return ObjectExpr2875()


def _arrow2880(get: IGetters) -> DataMap:
    def _arrow2879(__unit: None=None) -> Array[DataContext]:
        arg_1: Decoder_1[Array[DataContext]] = resize_array(decoder_1)
        object_arg: IRequiredGetter = get.Required
        return object_arg.Field("dataContexts", arg_1)

    return DataMap(_arrow2879())


decoder: Decoder_1[DataMap] = object(_arrow2880)

def encoder_compressed(string_table: Any, oa_table: Any, cell_table: Any, dm: DataMap) -> IEncodable:
    return encoder(dm)


def decoder_compressed(string_table: Array[str], oa_table: Array[OntologyAnnotation], cell_table: Array[CompositeCell]) -> Decoder_1[DataMap]:
    return decoder


__all__ = ["encoder", "decoder", "encoder_compressed", "decoder_compressed"]




from __future__ import annotations

import re
from math import ceil, floor, fmod
from typing import Any

from .types import FSharpRef
from .util import pad_left_and_right_with_zeros, pad_with_zeros


# TimeSpan is represented as an int which is the Tick value
# We can recompute everything from this value
class TimeSpan(int):
    pass


def create(
    days: float = 0,
    hours: float | None = None,
    minutes: float | None = None,
    seconds: float | None = None,
    milliseconds: float | None = None,
    microseconds: float | None = None,
) -> TimeSpan:
    match (days, hours, minutes, seconds, milliseconds, microseconds):
        # ticks constructor
        case (_, None, None, None, None, None):
            return TimeSpan(days)
        # hours, minutes, seconds constructor
        case (_, _, _, None, None, None):
            seconds = minutes
            minutes = hours
            hours = days
            days = 0
        # others constructor follows the correct order of arguments
        case _:
            pass

    return TimeSpan(
        days * 864000000000
        + (hours or 0) * 36000000000
        + (minutes or 0) * 600000000
        + (seconds or 0) * 10000000
        + (milliseconds or 0) * 10000
        + (microseconds or 0) * 10
    )


def total_nanoseconds(ts: TimeSpan) -> float:
    # We store timespan as the Tick value so nanoseconds step is 100
    return ts * 100


def total_microseconds(ts: TimeSpan) -> float:
    return ts / 10


def total_milliseconds(ts: TimeSpan) -> float:
    return ts / 10000


def total_seconds(ts: TimeSpan) -> float:
    return ts / 10000000


def total_minutes(ts: TimeSpan) -> float:
    return ts / 600000000


def total_hours(ts: TimeSpan) -> float:
    return ts / 36000000000


def total_days(ts: TimeSpan) -> float:
    return ts / 864000000000


def from_microseconds(micros: float) -> TimeSpan:
    return create(0, 0, 0, 0, 0, micros)


def from_milliseconds(msecs: int) -> TimeSpan:
    return create(0, 0, 0, 0, msecs)


def from_ticks(ticks: int) -> TimeSpan:
    return create(ticks)


def from_seconds(s: int) -> TimeSpan:
    return create(0, 0, s)


def from_minutes(m: int) -> TimeSpan:
    return create(0, m, 0)


def from_hours(h: int) -> TimeSpan:
    return create(h, 0, 0)


def from_days(d: int) -> TimeSpan:
    return create(d, 0, 0, 0)


def ticks(ts: TimeSpan) -> int:
    return int(ts)


def microseconds(ts: TimeSpan) -> int:
    return int(fmod(ts, 10000) / 10)


def milliseconds(ts: TimeSpan) -> int:
    return int(fmod(ts, 10000000) / 10000)


def seconds(ts: TimeSpan) -> int:
    return int(fmod(ts, 600000000) / 10000000)


def minutes(ts: TimeSpan) -> int:
    return int(fmod(ts, 36000000000) / 600000000)


def hours(ts: TimeSpan) -> int:
    return int(fmod(ts, 864000000000) / 36000000000)


def days(ts: TimeSpan) -> int:
    res = ts / 864000000000
    return ceil(res) if res < 0 else floor(res)


def negate(ts: TimeSpan) -> TimeSpan:
    return TimeSpan(-ts)


def duration(ts: TimeSpan) -> TimeSpan:
    return TimeSpan(abs(int(ts)))


def add(ts: TimeSpan, other: TimeSpan) -> TimeSpan:
    return TimeSpan(ts + other)


def subtract(ts: TimeSpan, other: TimeSpan) -> TimeSpan:
    return TimeSpan(ts - other)


def multiply(ts: TimeSpan, factor: float) -> TimeSpan:
    # We represents TimeSpan as a Tick which can't be a float
    # This also allows us
    return TimeSpan(int(ts * factor))


def divide(ts: TimeSpan, divisor: float) -> TimeSpan:
    return TimeSpan(int(ts / divisor))


def to_string(ts: TimeSpan, format: str = "c", _provider: Any | None = None) -> str:
    if format not in ["c", "g", "G"]:
        raise ValueError("Custom formats are not supported")

    d = abs(days(ts))
    h = abs(hours(ts))
    m = abs(minutes(ts))
    s = abs(seconds(ts))
    ms = abs(milliseconds(ts))
    sign: str = "-" if ts < 0 else ""

    ms_str = (
        ""
        if ms == 0 and format != "G"
        else "." + pad_left_and_right_with_zeros(ms, 3, 7)
        if format != "g"
        else "." + pad_with_zeros(ms, 3)
    )
    day_str = "" if d == 0 and format != "G" else f"{d}." if format == "c" else f"{d}:"
    hour_str = pad_with_zeros(h, 2) if format != "g" else str(h)

    return f"{sign}{day_str}{hour_str}:{pad_with_zeros(m, 2)}:{pad_with_zeros(s, 2)}{ms_str}"


_time_span_parse_regex = re.compile(
    r"^(-?)((\d+)\.)?(?:0*)([0-9]|0[0-9]|1[0-9]|2[0-3]):(?:0*)([0-5][0-9]|[0-9])(:(?:0*)([0-5][0-9]|[0-9]))?\.?(\d+)?$"
)


# Second argument is to not crash when using provides CultureInfo.InvariantCulture
def parse(string: str, _: Any | None = None) -> TimeSpan:
    first_dot = string.find(".")
    first_colon = string.find(":")
    if first_dot == -1 and first_colon == -1:
        # There is only a day ex: 4
        # parse as int
        try:
            d = int(string)
        except Exception:
            raise Exception("String '%s' was not recognized as a valid TimeSpan." % string)
        return from_days(d)
    if first_colon > 0:  # process time part
        r = _time_span_parse_regex.match(string)
        if r is not None and r.group(4) is not None and r.group(5) is not None:
            d = 0
            ms = 0
            s = 0
            sign = -1 if r.group(1) == "-" else 1
            h = int(r.group(4))
            m = int(r.group(5))
            if r.group(3) is not None:
                d = int(r.group(3))
            if r.group(7) is not None:
                s = int(r.group(7))
            if r.group(8) is not None:
                # Depending on the number of decimals passed, we need to adapt the numbers
                g_8: str = r.group(8)
                match len(g_8):
                    case 1:
                        ms = int(g_8) * 100
                    case 2:
                        ms = int(g_8) * 10
                    case 3:
                        ms = int(g_8)
                    case 4:
                        ms = int(g_8) / 10
                    case 5:
                        ms = int(g_8) / 100
                    case 6:
                        ms = int(g_8) / 1000
                    case 7:
                        ms = int(g_8) / 10000
                    case _:
                        raise Exception("String '%s' was not recognized as a valid TimeSpan." % string)
            return multiply(create(d, h, m, s, ms), sign)
    raise Exception("String '%s' was not recognized as a valid TimeSpan." % string)


def try_parse(
    string: str, def_value_or_format_provider: FSharpRef[TimeSpan] | Any, def_value: FSharpRef[TimeSpan] | None = None
) -> bool:
    # Find where the out_value is
    out_value: FSharpRef[TimeSpan] = def_value_or_format_provider

    # If we have 3 arguments, it means that the second argument is the format provider
    if def_value is not None:
        out_value = def_value

    try:
        out_value.contents = parse(string)
    except Exception:
        return False

    return True


__all__ = [
    "create",
    "total_microseconds",
    "total_milliseconds",
    "total_seconds",
    "total_minutes",
    "total_hours",
    "total_days",
    "from_ticks",
    "from_microseconds",
    "from_milliseconds",
    "from_hours",
    "from_minutes",
    "from_seconds",
    "negate",
    "duration",
    "total_seconds",
    "total_days",
    "total_minutes",
    "total_hours",
    "add",
    "subtract",
    "divide",
    "multiply",
    "parse",
    "try_parse",
]



from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ..arctrl_file_system.path import alternative_licensefile_names
from ..fable_library.array_ import equals_with
from ..fable_library.list import contains
from ..fable_library.types import Array
from ..fable_library.util import string_hash
from .contract import (Contract, DTOType, DTO)

def LicenseContractExtensions__007CLicensePath_007C__007C(input: Array[str]) -> str | None:
    (pattern_matching_result,) = (None,)
    def _arrow3714(x: str, y: str, input: Any=input) -> bool:
        return x == y

    if (len(input) == 1) if (not equals_with(_arrow3714, input, None)) else False:
        if input[0] == "LICENSE":
            pattern_matching_result = 0

        else: 
            class ObjectExpr3716:
                @property
                def Equals(self) -> Callable[[str, str], bool]:
                    def _arrow3715(x_1: str, y_1: str) -> bool:
                        return x_1 == y_1

                    return _arrow3715

                @property
                def GetHashCode(self) -> Callable[[str], int]:
                    return string_hash

            if contains(input[0], alternative_licensefile_names, ObjectExpr3716()):
                pattern_matching_result = 1

            else: 
                pattern_matching_result = 2



    else: 
        pattern_matching_result = 2

    if pattern_matching_result == 0:
        return "LICENSE"

    elif pattern_matching_result == 1:
        return input[0]

    elif pattern_matching_result == 2:
        return None



License_defaultLicenseContract: Contract = Contract.create_create("LICENSE", DTOType(10), DTO(1, "ALL RIGHTS RESERVED BY THE AUTHORS"))

__all__ = ["LicenseContractExtensions__007CLicensePath_007C__007C", "License_defaultLicenseContract"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ..fable_library.array_ import (try_find, exists, try_pick, head, tail, map, filter, equals_with, iterate, choose, append as append_1)
from ..fable_library.list import (cons, to_array as to_array_1, reverse, FSharpList, empty)
from ..fable_library.option import default_arg
from ..fable_library.reflection import (TypeInfo, string_type, array_type, union_type)
from ..fable_library.seq import (to_array, delay, append, singleton, empty as empty_1)
from ..fable_library.seq2 import (Array_groupBy, Array_distinct)
from ..fable_library.types import (Array, Union)
from ..fable_library.util import (IEnumerable_1, string_hash, array_hash, curry2)
from .path import (split, combine_many)

def _expr34() -> TypeInfo:
    return union_type("ARCtrl.FileSystem.FileSystemTree", [], FileSystemTree, lambda: [[("name", string_type)], [("name", string_type), ("children", array_type(FileSystemTree_reflection()))]])


class FileSystemTree(Union):
    def __init__(self, tag: int, *fields: Any) -> None:
        super().__init__()
        self.tag: int = tag or 0
        self.fields: Array[Any] = list(fields)

    @staticmethod
    def cases() -> list[str]:
        return ["File", "Folder"]

    @property
    def Name(self, __unit: None=None) -> str:
        this: FileSystemTree = self
        (pattern_matching_result, n) = (None, None)
        if this.tag == 1:
            pattern_matching_result = 0
            n = this.fields[0]

        else: 
            pattern_matching_result = 0
            n = this.fields[0]

        if pattern_matching_result == 0:
            return n


    @property
    def is_folder(self, __unit: None=None) -> bool:
        this: FileSystemTree = self
        return True if (this.tag == 1) else False

    @property
    def is_file(self, __unit: None=None) -> bool:
        this: FileSystemTree = self
        return True if (this.tag == 0) else False

    @staticmethod
    def ROOT_NAME() -> str:
        return "root"

    @staticmethod
    def create_file(name: str) -> FileSystemTree:
        return FileSystemTree(0, name)

    @staticmethod
    def create_folder(name: str, children: Array[FileSystemTree] | None=None) -> FileSystemTree:
        return FileSystemTree(1, name, default_arg(children, []))

    @staticmethod
    def create_root_folder(children: Array[FileSystemTree]) -> FileSystemTree:
        return FileSystemTree(1, FileSystemTree.ROOT_NAME(), children)

    def TryGetChildByName(self, name: str) -> FileSystemTree | None:
        this: FileSystemTree = self
        def predicate(c: FileSystemTree) -> bool:
            return c.Name == name

        return None if (this.tag == 0) else try_find(predicate, this.fields[1])

    @staticmethod
    def try_get_child_by_name(name: str) -> Callable[[FileSystemTree], FileSystemTree | None]:
        def _arrow6(fst: FileSystemTree) -> FileSystemTree | None:
            return fst.TryGetChildByName(name)

        return _arrow6

    def ContainsChildWithName(self, name: str) -> bool:
        this: FileSystemTree = self
        def predicate(c: FileSystemTree) -> bool:
            return c.Name == name

        return False if (this.tag == 0) else exists(predicate, this.fields[1])

    @staticmethod
    def contains_child_with_name(name: str) -> Callable[[FileSystemTree], bool]:
        def _arrow7(fst: FileSystemTree) -> bool:
            return fst.ContainsChildWithName(name)

        return _arrow7

    def TryGetPath(self, path: str) -> FileSystemTree | None:
        this: FileSystemTree = self
        def loop(parent: FileSystemTree, path_parts: Array[str]) -> FileSystemTree | None:
            if parent.tag == 0:
                if (path_parts[0] == parent.fields[0]) if (len(path_parts) == 1) else False:
                    return parent

                else: 
                    return None


            elif parent.fields[0] == "root":
                def chooser(child: FileSystemTree, parent: Any=parent, path_parts: Any=path_parts) -> FileSystemTree | None:
                    return loop(child, path_parts)

                return try_pick(chooser, parent.fields[1])

            elif parent.fields[0] == head(path_parts):
                if len(path_parts) == 1:
                    return parent

                else: 
                    remaining_parts: Array[str] = tail(path_parts)
                    def chooser_1(child_1: FileSystemTree, parent: Any=parent, path_parts: Any=path_parts) -> FileSystemTree | None:
                        return loop(child_1, remaining_parts)

                    return try_pick(chooser_1, parent.fields[1])


            else: 
                return None


        return loop(this, split(path))

    def AddFile(self, path: str) -> FileSystemTree:
        this: FileSystemTree = self
        existing_paths: Array[str] = this.ToFilePaths()
        def _arrow9(__unit: None=None) -> IEnumerable_1[str]:
            def _arrow8(__unit: None=None) -> IEnumerable_1[str]:
                return existing_paths

            return append(singleton(path), delay(_arrow8))

        file_paths: Array[str] = to_array(delay(_arrow9))
        return FileSystemTree.from_file_paths(file_paths)

    @staticmethod
    def add_file(path: str) -> Callable[[FileSystemTree], FileSystemTree]:
        def _arrow10(tree: FileSystemTree) -> FileSystemTree:
            return tree.AddFile(path)

        return _arrow10

    @staticmethod
    def from_file_paths(paths: Array[str]) -> FileSystemTree:
        def loop(paths_1: Array[Array[str]], parent: FileSystemTree) -> FileSystemTree:
            def mapping(arg: Array[str], paths_1: Any=paths_1, parent: Any=parent) -> FileSystemTree:
                name: str = head(arg)
                return FileSystemTree.create_file(name)

            def predicate(p: Array[str], paths_1: Any=paths_1, parent: Any=parent) -> bool:
                return len(p) == 1

            files: Array[FileSystemTree] = map(mapping, filter(predicate, paths_1), None)
            def mapping_2(tupled_arg: tuple[str, Array[Array[str]]], paths_1: Any=paths_1, parent: Any=parent) -> FileSystemTree:
                parent_1: FileSystemTree = FileSystemTree.create_folder(tupled_arg[0], [])
                def mapping_1(array_7: Array[str], tupled_arg: Any=tupled_arg) -> Array[str]:
                    return tail(array_7)

                return loop(map(mapping_1, tupled_arg[1], None), parent_1)

            def projection(x_2: Array[str], paths_1: Any=paths_1, parent: Any=parent) -> str:
                return head(x_2)

            def predicate_1(p_1: Array[str], paths_1: Any=paths_1, parent: Any=parent) -> bool:
                return len(p_1) > 1

            class ObjectExpr12:
                @property
                def Equals(self) -> Callable[[str, str], bool]:
                    def _arrow11(x_3: str, y_2: str) -> bool:
                        return x_3 == y_2

                    return _arrow11

                @property
                def GetHashCode(self) -> Callable[[str], int]:
                    return string_hash

            folders: Array[FileSystemTree] = map(mapping_2, Array_groupBy(projection, filter(predicate_1, paths_1), ObjectExpr12()), None)
            if parent.tag == 1:
                def _arrow14(__unit: None=None, paths_1: Any=paths_1, parent: Any=parent) -> IEnumerable_1[FileSystemTree]:
                    def _arrow13(__unit: None=None) -> IEnumerable_1[FileSystemTree]:
                        return folders

                    return append(files, delay(_arrow13))

                return FileSystemTree.create_folder(parent.fields[0], to_array(delay(_arrow14)))

            else: 
                return parent


        class ObjectExpr17:
            @property
            def Equals(self) -> Callable[[Array[str], Array[str]], bool]:
                def _arrow16(x: Array[str], y: Array[str]) -> bool:
                    def _arrow15(x_1: str, y_1: str) -> bool:
                        return x_1 == y_1

                    return equals_with(_arrow15, x, y)

                return _arrow16

            @property
            def GetHashCode(self) -> Callable[[Array[str]], int]:
                return array_hash

        return loop(Array_distinct(map(split, paths, None), ObjectExpr17()), FileSystemTree.create_folder(FileSystemTree.ROOT_NAME(), []))

    def ToFilePaths(self, remove_root: bool | None=None) -> Array[str]:
        this: FileSystemTree = self
        res: Array[str] = []
        def loop(output: FSharpList[str], parent: FileSystemTree) -> None:
            if parent.tag == 1:
                def action(filest: FileSystemTree, output: Any=output, parent: Any=parent) -> None:
                    loop(cons(parent.fields[0], output), filest)

                iterate(action, parent.fields[1])

            else: 
                item: str = combine_many(to_array_1(reverse(cons(parent.fields[0], output))))
                (res.append(item))


        if default_arg(remove_root, True):
            if this.tag == 0:
                (res.append(this.fields[0]))

            else: 
                iterate(curry2(loop)(empty()), this.fields[1])


        else: 
            loop(empty(), this)

        return list(res)

    @staticmethod
    def to_file_paths(remove_root: bool | None=None) -> Callable[[FileSystemTree], Array[str]]:
        def _arrow18(root: FileSystemTree) -> Array[str]:
            return root.ToFilePaths(remove_root)

        return _arrow18

    def FilterFiles(self, predicate: Callable[[str], bool]) -> FileSystemTree | None:
        this: FileSystemTree = self
        def loop(parent: FileSystemTree) -> FileSystemTree | None:
            if parent.tag == 1:
                return FileSystemTree(1, parent.fields[0], choose(loop, parent.fields[1], None))

            else: 
                n: str = parent.fields[0]
                if predicate(n):
                    return FileSystemTree(0, n)

                else: 
                    return None



        return loop(this)

    @staticmethod
    def filter_files(predicate: Callable[[str], bool]) -> Callable[[FileSystemTree], FileSystemTree | None]:
        def _arrow19(tree: FileSystemTree) -> FileSystemTree | None:
            return tree.FilterFiles(predicate)

        return _arrow19

    def FilterFolders(self, predicate: Callable[[str], bool]) -> FileSystemTree | None:
        this: FileSystemTree = self
        def loop(parent: FileSystemTree) -> FileSystemTree | None:
            if parent.tag == 1:
                n_1: str = parent.fields[0]
                if predicate(n_1):
                    return FileSystemTree(1, n_1, choose(loop, parent.fields[1], None))

                else: 
                    return None


            else: 
                return FileSystemTree(0, parent.fields[0])


        return loop(this)

    @staticmethod
    def filter_folders(predicate: Callable[[str], bool]) -> Callable[[FileSystemTree], FileSystemTree | None]:
        def _arrow20(tree: FileSystemTree) -> FileSystemTree | None:
            return tree.FilterFolders(predicate)

        return _arrow20

    def Filter(self, predicate: Callable[[str], bool]) -> FileSystemTree | None:
        this: FileSystemTree = self
        def loop(parent: FileSystemTree) -> FileSystemTree | None:
            if parent.tag == 1:
                n_1: str = parent.fields[0]
                if predicate(n_1):
                    filtered_children: Array[FileSystemTree] = choose(loop, parent.fields[1], None)
                    if len(filtered_children) == 0:
                        return None

                    else: 
                        return FileSystemTree(1, n_1, filtered_children)


                else: 
                    return None


            else: 
                n: str = parent.fields[0]
                if predicate(n):
                    return FileSystemTree(0, n)

                else: 
                    return None



        return loop(this)

    @staticmethod
    def filter(predicate: Callable[[str], bool]) -> Callable[[FileSystemTree], FileSystemTree | None]:
        def _arrow21(tree: FileSystemTree) -> FileSystemTree | None:
            return tree.Filter(predicate)

        return _arrow21

    def Union(self, other_fst: FileSystemTree) -> FileSystemTree:
        this: FileSystemTree = self
        def _arrow22(__unit: None=None) -> Array[str]:
            array_1: Array[str] = this.ToFilePaths()
            return append_1(other_fst.ToFilePaths(), array_1, None)

        class ObjectExpr24:
            @property
            def Equals(self) -> Callable[[str, str], bool]:
                def _arrow23(x: str, y: str) -> bool:
                    return x == y

                return _arrow23

            @property
            def GetHashCode(self) -> Callable[[str], int]:
                return string_hash

        paths: Array[str] = Array_distinct(_arrow22(), ObjectExpr24())
        return FileSystemTree.from_file_paths(paths)

    def Copy(self, __unit: None=None) -> FileSystemTree:
        this: FileSystemTree = self
        def mapping(c: FileSystemTree) -> FileSystemTree:
            return c.Copy()

        return FileSystemTree(0, this.fields[0]) if (this.tag == 0) else FileSystemTree(1, this.fields[0], map(mapping, this.fields[1], None))

    @staticmethod
    def create_git_keep_file(__unit: None=None) -> FileSystemTree:
        return FileSystemTree.create_file(".gitkeep")

    @staticmethod
    def create_readme_file(__unit: None=None) -> FileSystemTree:
        return FileSystemTree.create_file("README.md")

    @staticmethod
    def create_empty_folder(name: str) -> FileSystemTree:
        return FileSystemTree.create_folder(name, [FileSystemTree.create_git_keep_file()])

    @staticmethod
    def create_assay_folder(assay_name: str, has_data_map: bool | None=None) -> FileSystemTree:
        has_data_map_1: bool = default_arg(has_data_map, False)
        dataset: FileSystemTree = FileSystemTree.create_empty_folder("dataset")
        protocols: FileSystemTree = FileSystemTree.create_empty_folder("protocols")
        readme: FileSystemTree = FileSystemTree.create_readme_file()
        assay_file: FileSystemTree = FileSystemTree.create_file("isa.assay.xlsx")
        if has_data_map_1:
            data_map_file: FileSystemTree = FileSystemTree.create_file("isa.datamap.xlsx")
            return FileSystemTree.create_folder(assay_name, [dataset, protocols, assay_file, readme, data_map_file])

        else: 
            return FileSystemTree.create_folder(assay_name, [dataset, protocols, assay_file, readme])


    @staticmethod
    def create_study_folder(study_name: str, has_data_map: bool | None=None) -> FileSystemTree:
        has_data_map_1: bool = default_arg(has_data_map, False)
        resources: FileSystemTree = FileSystemTree.create_empty_folder("resources")
        protocols: FileSystemTree = FileSystemTree.create_empty_folder("protocols")
        readme: FileSystemTree = FileSystemTree.create_readme_file()
        study_file: FileSystemTree = FileSystemTree.create_file("isa.study.xlsx")
        if has_data_map_1:
            data_map_file: FileSystemTree = FileSystemTree.create_file("isa.datamap.xlsx")
            return FileSystemTree.create_folder(study_name, [resources, protocols, study_file, readme, data_map_file])

        else: 
            return FileSystemTree.create_folder(study_name, [resources, protocols, study_file, readme])


    @staticmethod
    def create_workflow_folder(workflow_name: str, has_cwl: bool | None=None, has_data_map: bool | None=None) -> FileSystemTree:
        has_data_map_1: bool = default_arg(has_data_map, False)
        has_cwl_1: bool = default_arg(has_cwl, False)
        readme: FileSystemTree = FileSystemTree.create_readme_file()
        workflow_file: FileSystemTree = FileSystemTree.create_file("isa.workflow.xlsx")
        def _arrow28(__unit: None=None) -> IEnumerable_1[FileSystemTree]:
            def _arrow27(__unit: None=None) -> IEnumerable_1[FileSystemTree]:
                def _arrow26(__unit: None=None) -> IEnumerable_1[FileSystemTree]:
                    def _arrow25(__unit: None=None) -> IEnumerable_1[FileSystemTree]:
                        return singleton(FileSystemTree.create_file("isa.datamap.xlsx")) if has_data_map_1 else empty_1()

                    return append(singleton(FileSystemTree.create_file("workflow.cwl")) if has_cwl_1 else empty_1(), delay(_arrow25))

                return append(singleton(readme), delay(_arrow26))

            return append(singleton(workflow_file), delay(_arrow27))

        return FileSystemTree.create_folder(workflow_name, to_array(delay(_arrow28)))

    @staticmethod
    def create_run_folder(run_name: str, has_cwl: bool | None=None, has_yml: bool | None=None, has_data_map: bool | None=None) -> FileSystemTree:
        has_data_map_1: bool = default_arg(has_data_map, False)
        has_cwl_1: bool = default_arg(has_cwl, False)
        has_yml_1: bool = default_arg(has_yml, False)
        readme: FileSystemTree = FileSystemTree.create_readme_file()
        run_file: FileSystemTree = FileSystemTree.create_file("isa.run.xlsx")
        def _arrow33(__unit: None=None) -> IEnumerable_1[FileSystemTree]:
            def _arrow32(__unit: None=None) -> IEnumerable_1[FileSystemTree]:
                def _arrow31(__unit: None=None) -> IEnumerable_1[FileSystemTree]:
                    def _arrow30(__unit: None=None) -> IEnumerable_1[FileSystemTree]:
                        def _arrow29(__unit: None=None) -> IEnumerable_1[FileSystemTree]:
                            return singleton(FileSystemTree.create_file("isa.datamap.xlsx")) if has_data_map_1 else empty_1()

                        return append(singleton(FileSystemTree.create_file("run.yml")) if has_yml_1 else empty_1(), delay(_arrow29))

                    return append(singleton(FileSystemTree.create_file("run.cwl")) if has_cwl_1 else empty_1(), delay(_arrow30))

                return append(singleton(readme), delay(_arrow31))

            return append(singleton(run_file), delay(_arrow32))

        return FileSystemTree.create_folder(run_name, to_array(delay(_arrow33)))

    @staticmethod
    def create_investigation_file(__unit: None=None) -> FileSystemTree:
        return FileSystemTree.create_file("isa.investigation.xlsx")

    @staticmethod
    def create_license_file(__unit: None=None) -> FileSystemTree:
        return FileSystemTree.create_file("LICENSE")

    @staticmethod
    def create_assays_folder(assays: Array[FileSystemTree]) -> FileSystemTree:
        return FileSystemTree.create_folder("assays", append_1([FileSystemTree.create_git_keep_file()], assays, None))

    @staticmethod
    def create_studies_folder(studies: Array[FileSystemTree]) -> FileSystemTree:
        return FileSystemTree.create_folder("studies", append_1([FileSystemTree.create_git_keep_file()], studies, None))

    @staticmethod
    def create_workflows_folder(workflows: Array[FileSystemTree]) -> FileSystemTree:
        return FileSystemTree.create_folder("workflows", append_1([FileSystemTree.create_git_keep_file()], workflows, None))

    @staticmethod
    def create_runs_folder(runs: Array[FileSystemTree]) -> FileSystemTree:
        return FileSystemTree.create_folder("runs", append_1([FileSystemTree.create_git_keep_file()], runs, None))


FileSystemTree_reflection = _expr34

__all__ = ["FileSystemTree_reflection"]




from __future__ import annotations

from collections.abc import Callable
from typing import Any, TypeVar


_T = TypeVar("_T")


def exists(predicate: Callable[[_T], bool], xs: list[_T]) -> bool:
    """Test if a predicate is true for at least one element in a list."""

    for x in xs:
        if predicate(x):
            return True
    return False


def find_index(predicate: Callable[[_T], bool], xs: list[_T]) -> int:
    """Find the index of the first element in a list that satisfies a predicate."""

    for i, x in enumerate(xs):
        if predicate(x):
            return i
    return -1


def remove_range(start: int, count: int, xs: list[Any]) -> None:
    """Remove a range of elements from a list in-place."""

    del xs[start : start + count]



from __future__ import annotations
from typing import Any
from ...fable_library.option import (some, default_arg)
from ...fable_library.reflection import (TypeInfo, class_type)
from ...fable_library.seq import skip
from ...fable_library.string_ import (to_fail, printf)
from ...fable_library.util import (equals, IEnumerable_1)
from ..Cells.fs_cell import FsCell
from ..Cells.fs_cells_collection import FsCellsCollection
from ..fs_address import FsAddress
from ..Ranges.fs_range_address import FsRangeAddress
from ..Ranges.fs_range_column import FsRangeColumn

def _expr235() -> TypeInfo:
    return class_type("FsSpreadsheet.FsTableField", None, FsTableField)


class FsTableField:
    def __init__(self, name: str, index: int | None=None, column: FsRangeColumn | None=None, totals_row_label: Any | None=None, totals_row_function: Any | None=None) -> None:
        self._totalsRowsFunction: Any = default_arg(totals_row_function, None)
        self._totalsRowLabel: Any = default_arg(totals_row_label, None)
        self._column: FsRangeColumn = default_arg(column, None)
        self._index: int = default_arg(index, 0) or 0
        self._name: str = name
        self._Column: FsRangeColumn = self._column

    @property
    def Column(self, __unit: None=None) -> FsRangeColumn:
        __: FsTableField = self
        return __._Column

    @Column.setter
    def Column(self, v: FsRangeColumn) -> None:
        __: FsTableField = self
        __._Column = v

    @property
    def Index(self, __unit: None=None) -> int:
        this: FsTableField = self
        return this._index

    @Index.setter
    def Index(self, index: int) -> None:
        this: FsTableField = self
        if index == this._index:
            pass

        else: 
            this._index = index or 0
            if equals(this._column, None):
                pass

            else: 
                ind_diff: int = (index - this._index) or 0
                new_col: FsRangeColumn = FsRangeColumn(FsRangeAddress(FsAddress(this.Column.RangeAddress.FirstAddress.RowNumber, this._index + ind_diff), FsAddress(this.Column.RangeAddress.LastAddress.RowNumber, this._index + ind_diff)))
                this.Column = new_col



    @property
    def Name(self, __unit: None=None) -> str:
        this: FsTableField = self
        return this._name

    def SetName(self, name: str, cells_collection: FsCellsCollection, show_header_row: bool) -> None:
        this: FsTableField = self
        this._name = name
        if show_header_row:
            this.Column.FirstCell(cells_collection).SetValueAs(name)


    @staticmethod
    def set_name(name: str, cells_collection: FsCellsCollection, show_header_row: bool, table_field: FsTableField) -> FsTableField:
        table_field.SetName(name, cells_collection, show_header_row)
        return table_field

    def Copy(self, __unit: None=None) -> FsTableField:
        this: FsTableField = self
        col: FsRangeColumn = this.Column.Copy()
        ind: int = this.Index or 0
        return FsTableField(this.Name, ind, col, some(None), some(None))

    @staticmethod
    def copy(table_field: FsTableField) -> FsTableField:
        return table_field.Copy()

    def HeaderCell(self, cells_collection: FsCellsCollection, show_header_row: bool) -> FsCell:
        this: FsTableField = self
        if not show_header_row:
            arg: str = this._name
            return to_fail(printf("tried to get header cell of table field \"%s\" even though showHeaderRow is set to zero"))(arg)

        else: 
            return this.Column.FirstCell(cells_collection)


    @staticmethod
    def get_header_cell(cells_collection: FsCellsCollection, show_header_row: bool, table_field: FsTableField) -> FsCell:
        return table_field.HeaderCell(cells_collection, show_header_row)

    def DataCells(self, cells_collection: FsCellsCollection) -> IEnumerable_1[FsCell]:
        this: FsTableField = self
        return skip(1, this.Column.Cells(cells_collection))

    @staticmethod
    def get_data_cells(cells_collection: FsCellsCollection, table_field: FsTableField) -> IEnumerable_1[FsCell]:
        return table_field.DataCells(cells_collection)


FsTableField_reflection = _expr235

def FsTableField__ctor_E675DBB(name: str, index: int | None=None, column: FsRangeColumn | None=None, totals_row_label: Any | None=None, totals_row_function: Any | None=None) -> FsTableField:
    return FsTableField(name, index, column, totals_row_label, totals_row_function)


__all__ = ["FsTableField_reflection"]




from __future__ import annotations
from abc import abstractmethod
from collections.abc import Callable
from datetime import datetime
from typing import (Any, Protocol, TypeVar)
from ..fable_library.array_ import fold
from ..fable_library.date import to_universal_time
from ..fable_library.list import (FSharpList, is_empty, length, head, empty, cons, tail)
from ..fable_library.map_util import add_to_dict
from ..fable_library.mutable_map import Dictionary
from ..fable_library.result import (FSharpResult_2, Result_Map)
from ..fable_library.seq import exists
from ..fable_library.set import (contains, of_seq)
from ..fable_library.string_ import (to_text, printf)
from ..fable_library.types import Array
from ..fable_library.util import (compare_primitives, IEnumerable_1, int32_to_string, equals, structural_hash)
from ..thoth_json_core.decode import (string, Getters_2__ctor_Z4BE6C149, Getters_2, Getters_2__get_Errors, IGetters, map, datetime_local, tuple2, int_1)
from ..thoth_json_core.types import (ErrorReason_1, Decoder_1, IDecoderHelpers_1)

_JSONVALUE = TypeVar("_JSONVALUE")

__A_ = TypeVar("__A_")

_VALUE_ = TypeVar("_VALUE_")

_VALUE = TypeVar("_VALUE")

_KEY__ = TypeVar("_KEY__")

_KEY_ = TypeVar("_KEY_")

class DateTimeStatic(Protocol):
    @abstractmethod
    def from_time_stamp(self, timestamp: float) -> Any:
        ...


def PyTime_toUniversalTimePy(dt: Any) -> Any:
    timestamp: float = to_universal_time(dt).timestamp()
    return datetime.fromtimestamp(timestamp=timestamp)


def Helpers_prependPath(path: str, err_: str, err__1: ErrorReason_1[Any]) -> tuple[str, ErrorReason_1[_JSONVALUE]]:
    err: tuple[str, ErrorReason_1[_JSONVALUE]] = (err_, err__1)
    return (path + err[0], err[1])


def Decode_isURI(s: str) -> bool:
    return True


class ObjectExpr1862(Decoder_1[str]):
    def Decode(self, s: IDecoderHelpers_1[Any], json: Any) -> FSharpResult_2[str, tuple[str, ErrorReason_1[__A_]]]:
        match_value: FSharpResult_2[str, tuple[str, ErrorReason_1[__A_]]] = string.Decode(s, json)
        if match_value.tag == 1:
            return FSharpResult_2(1, match_value.fields[0])

        elif Decode_isURI(match_value.fields[0]):
            return FSharpResult_2(0, match_value.fields[0])

        else: 
            s_3: str = match_value.fields[0]
            return FSharpResult_2(1, (s_3, ErrorReason_1(6, to_text(printf("Expected URI, got %s"))(s_3))))



Decode_uri: Decoder_1[str] = ObjectExpr1862()

def Decode_hasUnknownFields(helpers: IDecoderHelpers_1[Any], known_fields: Any, json: Any) -> bool:
    def predicate(x: str, helpers: Any=helpers, known_fields: Any=known_fields, json: Any=json) -> bool:
        return not contains(x, known_fields)

    return exists(predicate, helpers.get_properties(json))


def Decode_objectNoAdditionalProperties(allowed_properties: IEnumerable_1[str], builder: Callable[[IGetters], _VALUE]) -> Decoder_1[Any]:
    class ObjectExpr1863:
        @property
        def Compare(self) -> Callable[[str, str], int]:
            return compare_primitives

    allowed_properties_1: Any = of_seq(allowed_properties, ObjectExpr1863())
    class ObjectExpr1864(Decoder_1[_VALUE_]):
        def Decode(self, helpers: IDecoderHelpers_1[Any], value: Any, allowed_properties: Any=allowed_properties, builder: Any=builder) -> FSharpResult_2[_VALUE_, tuple[str, ErrorReason_1[__A_]]]:
            getters: Getters_2[__A_, Any] = Getters_2__ctor_Z4BE6C149(helpers, value)
            if Decode_hasUnknownFields(helpers, allowed_properties_1, value):
                return FSharpResult_2(1, ("Unknown fields in object", ErrorReason_1(0, "", value)))

            else: 
                result: _VALUE_ = builder(getters)
                match_value: FSharpList[tuple[str, ErrorReason_1[__A_]]] = Getters_2__get_Errors(getters)
                if not is_empty(match_value):
                    errors: FSharpList[tuple[str, ErrorReason_1[__A_]]] = match_value
                    return FSharpResult_2(1, ("", ErrorReason_1(7, errors))) if (length(errors) > 1) else FSharpResult_2(1, head(match_value))

                else: 
                    return FSharpResult_2(0, result)



    return ObjectExpr1864()


def Decode_noAdditionalProperties(allowed_properties: IEnumerable_1[str], decoder: Decoder_1[Any]) -> Decoder_1[Any]:
    class ObjectExpr1865:
        @property
        def Compare(self) -> Callable[[str, str], int]:
            return compare_primitives

    allowed_properties_1: Any = of_seq(allowed_properties, ObjectExpr1865())
    class ObjectExpr1866(Decoder_1[_VALUE_]):
        def Decode(self, helpers: IDecoderHelpers_1[Any], value: Any, allowed_properties: Any=allowed_properties, decoder: Any=decoder) -> FSharpResult_2[_VALUE_, tuple[str, ErrorReason_1[__A_]]]:
            getters: Getters_2[__A_, Any] = Getters_2__ctor_Z4BE6C149(helpers, value)
            return FSharpResult_2(1, ("Unknown fields in object", ErrorReason_1(0, "", value))) if Decode_hasUnknownFields(helpers, allowed_properties_1, value) else decoder.Decode(helpers, value)

    return ObjectExpr1866()


def Decode_resizeArrayOrSingleton(decoder: Decoder_1[Any]) -> Decoder_1[Array[Any]]:
    class ObjectExpr1868(Decoder_1[Array[_VALUE_]]):
        def Decode(self, helpers: IDecoderHelpers_1[Any], value: Any, decoder: Any=decoder) -> FSharpResult_2[Array[_VALUE_], tuple[str, ErrorReason_1[__A_]]]:
            if helpers.is_array(value):
                i: int = -1
                def folder(acc: FSharpResult_2[Array[_VALUE_], tuple[str, ErrorReason_1[__A_]]], value_1: __A_) -> FSharpResult_2[Array[_VALUE_], tuple[str, ErrorReason_1[__A_]]]:
                    nonlocal i
                    i = (i + 1) or 0
                    if acc.tag == 0:
                        acc_1: Array[_VALUE_] = acc.fields[0]
                        match_value: FSharpResult_2[_VALUE_, tuple[str, ErrorReason_1[__A_]]] = decoder.Decode(helpers, value_1)
                        if match_value.tag == 0:
                            (acc_1.append(match_value.fields[0]))
                            return FSharpResult_2(0, acc_1)

                        else: 
                            def _arrow1867(__unit: None=None, acc: Any=acc, value_1: Any=value_1) -> tuple[str, ErrorReason_1[__A_]]:
                                tupled_arg: tuple[str, ErrorReason_1[__A_]] = match_value.fields[0]
                                return Helpers_prependPath((".[" + int32_to_string(i)) + "]", tupled_arg[0], tupled_arg[1])

                            return FSharpResult_2(1, _arrow1867())


                    else: 
                        return acc


                return fold(folder, FSharpResult_2(0, []), helpers.as_array(value))

            else: 
                def mapping(x: _VALUE_ | None=None) -> Array[_VALUE_]:
                    return [x]

                return Result_Map(mapping, decoder.Decode(helpers, value))


    return ObjectExpr1868()


Decode_datetime: Decoder_1[Any] = map(PyTime_toUniversalTimePy, datetime_local)

def Decode_dictionary(key_decoder: Decoder_1[Any], value_decoder: Decoder_1[Any]) -> Decoder_1[Any]:
    class ObjectExpr1870(Decoder_1[Any]):
        def Decode(self, helpers: IDecoderHelpers_1[Any], value: Any, key_decoder: Any=key_decoder, value_decoder: Any=value_decoder) -> FSharpResult_2[Any, tuple[str, ErrorReason_1[__A_]]]:
            if helpers.is_array(value):
                errors: FSharpList[Any] = empty()
                tokens: Array[__A_] = helpers.as_array(value)
                class ObjectExpr1869:
                    @property
                    def Equals(self) -> Callable[[_KEY__, _KEY__], bool]:
                        return equals

                    @property
                    def GetHashCode(self) -> Callable[[_KEY__], int]:
                        return structural_hash

                dict_1: Any = Dictionary([], ObjectExpr1869())
                decoder: Decoder_1[tuple[_KEY_, _VALUE_]] = tuple2(key_decoder, value_decoder)
                def folder(acc: FSharpResult_2[Any, tuple[str, ErrorReason_1[__A_]]], value_1: __A_) -> FSharpResult_2[Any, tuple[str, ErrorReason_1[__A_]]]:
                    if acc.tag == 0:
                        acc_1: Any = acc.fields[0]
                        match_value: FSharpResult_2[tuple[_KEY_, _VALUE_], tuple[str, ErrorReason_1[__A_]]] = decoder.Decode(helpers, value_1)
                        if match_value.tag == 0:
                            tupled_arg: tuple[_KEY_, _VALUE_] = match_value.fields[0]
                            add_to_dict(acc_1, tupled_arg[0], tupled_arg[1])
                            return FSharpResult_2(0, acc_1)

                        else: 
                            return FSharpResult_2(1, match_value.fields[0])


                    else: 
                        return acc


                return fold(folder, FSharpResult_2(0, dict_1), tokens)

            else: 
                return FSharpResult_2(1, ("", ErrorReason_1(0, "an object", value)))


    return ObjectExpr1870()


def Decode_intDictionary(value_decoder: Decoder_1[Any]) -> Decoder_1[Any]:
    class ObjectExpr1871(Decoder_1[Any]):
        def Decode(self, helpers: IDecoderHelpers_1[Any], value: Any, value_decoder: Any=value_decoder) -> FSharpResult_2[Any, tuple[str, ErrorReason_1[__A_]]]:
            if helpers.is_array(value):
                errors: FSharpList[Any] = empty()
                tokens: Array[__A_] = helpers.as_array(value)
                dict_1: Any = dict([])
                decoder: Decoder_1[tuple[int, _VALUE_]] = tuple2(int_1, value_decoder)
                def folder(acc: FSharpResult_2[Any, tuple[str, ErrorReason_1[__A_]]], value_1: __A_) -> FSharpResult_2[Any, tuple[str, ErrorReason_1[__A_]]]:
                    if acc.tag == 0:
                        acc_1: Any = acc.fields[0]
                        match_value: FSharpResult_2[tuple[int, _VALUE_], tuple[str, ErrorReason_1[__A_]]] = decoder.Decode(helpers, value_1)
                        if match_value.tag == 0:
                            tupled_arg: tuple[int, _VALUE_] = match_value.fields[0]
                            add_to_dict(acc_1, tupled_arg[0], tupled_arg[1])
                            return FSharpResult_2(0, acc_1)

                        else: 
                            return FSharpResult_2(1, match_value.fields[0])


                    else: 
                        return acc


                return fold(folder, FSharpResult_2(0, dict_1), tokens)

            else: 
                return FSharpResult_2(1, ("", ErrorReason_1(0, "an object", value)))


    return ObjectExpr1871()


def Decode_tryOneOf(decoders: FSharpList[Decoder_1[Any]]) -> Decoder_1[Any]:
    class ObjectExpr1872(Decoder_1[_VALUE_]):
        def Decode(self, helpers: IDecoderHelpers_1[Any], value: Any, decoders: Any=decoders) -> FSharpResult_2[_VALUE_, tuple[str, ErrorReason_1[__A_]]]:
            def loop(errors_mut: FSharpList[tuple[str, ErrorReason_1[__A_]]], decoders_1_mut: FSharpList[Decoder_1[_VALUE_]]) -> FSharpResult_2[_VALUE_, tuple[str, ErrorReason_1[__A_]]]:
                while True:
                    (errors, decoders_1) = (errors_mut, decoders_1_mut)
                    if not is_empty(decoders_1):
                        decoding_result: FSharpResult_2[_VALUE_, tuple[str, ErrorReason_1[__A_]]]
                        try: 
                            decoding_result = head(decoders_1).Decode(helpers, value)

                        except Exception as e:
                            decoding_result = FSharpResult_2(1, ("", ErrorReason_1(6, str(e))))

                        if decoding_result.tag == 1:
                            errors_mut = cons(decoding_result.fields[0], errors)
                            decoders_1_mut = tail(decoders_1)
                            continue

                        else: 
                            return FSharpResult_2(0, decoding_result.fields[0])


                    else: 
                        return FSharpResult_2(1, ("", ErrorReason_1(7, errors)))

                    break

            return loop(empty(), decoders)

    return ObjectExpr1872()


__all__ = ["PyTime_toUniversalTimePy", "Helpers_prependPath", "Decode_isURI", "Decode_uri", "Decode_hasUnknownFields", "Decode_objectNoAdditionalProperties", "Decode_noAdditionalProperties", "Decode_resizeArrayOrSingleton", "Decode_datetime", "Decode_dictionary", "Decode_intDictionary", "Decode_tryOneOf"]




from __future__ import annotations
from typing import Any
from ..fable_library.option import default_arg
from ..fable_library.reflection import (TypeInfo, class_type)
from ..fable_library.types import Array
from ..fable_library.util import safe_hash
from .comment import Comment
from .Helper.collections_ import ResizeArray_map
from .Helper.hash_codes import (box_hash_array, box_hash_option, box_hash_seq)

def _expr740() -> TypeInfo:
    return class_type("ARCtrl.OntologySourceReference", None, OntologySourceReference)


class OntologySourceReference:
    def __init__(self, description: str | None=None, file: str | None=None, name: str | None=None, version: str | None=None, comments: Array[Comment] | None=None) -> None:
        self._description: str | None = description
        self._file: str | None = file
        self._name: str | None = name
        self._version: str | None = version
        self._comments: Array[Comment] = default_arg(comments, [])

    @property
    def Description(self, __unit: None=None) -> str | None:
        this: OntologySourceReference = self
        return this._description

    @Description.setter
    def Description(self, description: str | None=None) -> None:
        this: OntologySourceReference = self
        this._description = description

    @property
    def File(self, __unit: None=None) -> str | None:
        this: OntologySourceReference = self
        return this._file

    @File.setter
    def File(self, file: str | None=None) -> None:
        this: OntologySourceReference = self
        this._file = file

    @property
    def Name(self, __unit: None=None) -> str | None:
        this: OntologySourceReference = self
        return this._name

    @Name.setter
    def Name(self, name: str | None=None) -> None:
        this: OntologySourceReference = self
        this._name = name

    @property
    def Version(self, __unit: None=None) -> str | None:
        this: OntologySourceReference = self
        return this._version

    @Version.setter
    def Version(self, version: str | None=None) -> None:
        this: OntologySourceReference = self
        this._version = version

    @property
    def Comments(self, __unit: None=None) -> Array[Comment]:
        this: OntologySourceReference = self
        return this._comments

    @Comments.setter
    def Comments(self, comments: Array[Comment]) -> None:
        this: OntologySourceReference = self
        this._comments = comments

    @staticmethod
    def make(description: str | None, file: str | None, name: str | None, version: str | None, comments: Array[Comment]) -> OntologySourceReference:
        return OntologySourceReference(description, file, name, version, comments)

    @staticmethod
    def create(description: str | None=None, file: str | None=None, name: str | None=None, version: str | None=None, comments: Array[Comment] | None=None) -> OntologySourceReference:
        comments_1: Array[Comment] = default_arg(comments, [])
        return OntologySourceReference.make(description, file, name, version, comments_1)

    @staticmethod
    def empty() -> OntologySourceReference:
        return OntologySourceReference.create()

    def Copy(self, __unit: None=None) -> OntologySourceReference:
        this: OntologySourceReference = self
        def f(c: Comment) -> Comment:
            return c.Copy()

        next_comments: Array[Comment] = ResizeArray_map(f, this.Comments)
        description: str | None = this.Description
        file: str | None = this.File
        name: str | None = this.Name
        version: str | None = this.Version
        return OntologySourceReference.make(description, file, name, version, next_comments)

    def __hash__(self, __unit: None=None) -> Any:
        this: OntologySourceReference = self
        return box_hash_array([box_hash_option(this.Description), box_hash_option(this.File), box_hash_option(this.Name), box_hash_option(this.Version), box_hash_seq(this.Comments)])

    def __eq__(self, obj: Any=None) -> bool:
        this: OntologySourceReference = self
        return (safe_hash(this) == safe_hash(obj)) if isinstance(obj, OntologySourceReference) else False


OntologySourceReference_reflection = _expr740

def OntologySourceReference__ctor_7C9A7CF8(description: str | None=None, file: str | None=None, name: str | None=None, version: str | None=None, comments: Array[Comment] | None=None) -> OntologySourceReference:
    return OntologySourceReference(description, file, name, version, comments)


__all__ = ["OntologySourceReference_reflection"]




from __future__ import annotations
from ...fable_library.reflection import (TypeInfo, class_type)
from ...fable_library.util import IEnumerable_1
from ..Cells.fs_cell import FsCell
from ..Cells.fs_cells_collection import FsCellsCollection
from ..fs_address import FsAddress
from .fs_range_address import FsRangeAddress
from .fs_range_base import (FsRangeBase, FsRangeBase_reflection)

def _expr198() -> TypeInfo:
    return class_type("FsSpreadsheet.FsRangeRow", None, FsRangeRow, FsRangeBase_reflection())


class FsRangeRow(FsRangeBase):
    def __init__(self, range_address: FsRangeAddress) -> None:
        super().__init__(range_address)
        pass

    @staticmethod
    def from_index(index: int) -> FsRangeRow:
        return FsRangeRow(FsRangeAddress(FsAddress(index, 0), FsAddress(index, 0)))

    @property
    def Index(self, __unit: None=None) -> int:
        self_1: FsRangeRow = self
        return self_1.RangeAddress.FirstAddress.RowNumber

    @Index.setter
    def Index(self, i: int) -> None:
        self_1: FsRangeRow = self
        self_1.RangeAddress.FirstAddress.RowNumber = i or 0
        self_1.RangeAddress.LastAddress.RowNumber = i or 0

    def Cell(self, column_index: int, cells: FsCellsCollection) -> FsCell:
        return super().Cell(FsAddress(1, (column_index - super().RangeAddress.FirstAddress.ColumnNumber) + 1), cells)

    def Cells(self, cells: FsCellsCollection) -> IEnumerable_1[FsCell]:
        return super().Cells(cells)


FsRangeRow_reflection = _expr198

def FsRangeRow__ctor_6A2513BC(range_address: FsRangeAddress) -> FsRangeRow:
    return FsRangeRow(range_address)


__all__ = ["FsRangeRow_reflection"]




from __future__ import annotations
from collections.abc import Callable
from dataclasses import dataclass
from typing import Any
from ...fable_library.list import (try_find, FSharpList, exists, append, singleton, filter)
from ...fable_library.option import (default_arg, map)
from ...fable_library.reflection import (TypeInfo, string_type, option_type, array_type, record_type)
from ...fable_library.types import (Array, to_string, Record)
from ...fable_library.util import equals
from ..comment import (Comment, Comment_reflection)
from ..Helper.collections_ import (Option_fromValueWithDefault, ResizeArray_map)
from ..ontology_annotation import (OntologyAnnotation, OntologyAnnotation_reflection)

def _expr745() -> TypeInfo:
    return record_type("ARCtrl.Process.Factor", [], Factor, lambda: [("Name", option_type(string_type)), ("FactorType", option_type(OntologyAnnotation_reflection())), ("Comments", option_type(array_type(Comment_reflection())))])


@dataclass(eq = False, repr = False, slots = True)
class Factor(Record):
    Name: str | None
    FactorType: OntologyAnnotation | None
    Comments: Array[Comment] | None
    @staticmethod
    def make(name: str | None=None, factor_type: OntologyAnnotation | None=None, comments: Array[Comment] | None=None) -> Factor:
        return Factor(name, factor_type, comments)

    @staticmethod
    def create(Name: str | None=None, FactorType: OntologyAnnotation | None=None, Comments: Array[Comment] | None=None) -> Factor:
        return Factor.make(Name, FactorType, Comments)

    @staticmethod
    def empty() -> Factor:
        return Factor.create()

    @staticmethod
    def from_string(name: str, term: str, source: str, accession: str, comments: Array[Comment] | None=None) -> Factor:
        oa: OntologyAnnotation = OntologyAnnotation.create(term, source, accession, comments)
        name_1: str | None = Option_fromValueWithDefault("", name)
        factor_type: OntologyAnnotation | None = Option_fromValueWithDefault(OntologyAnnotation(), oa)
        return Factor.make(name_1, factor_type, None)

    @staticmethod
    def to_string_object(factor: Factor) -> dict[str, Any]:
        value: dict[str, Any] = {
            "TermAccessionNumber": "",
            "TermName": "",
            "TermSourceREF": ""
        }
        def mapping(oa: OntologyAnnotation) -> dict[str, Any]:
            return OntologyAnnotation.to_string_object(oa)

        return default_arg(map(mapping, factor.FactorType), value)

    @property
    def NameText(self, __unit: None=None) -> str:
        this: Factor = self
        return default_arg(this.Name, "")

    def MapCategory(self, f: Callable[[OntologyAnnotation], OntologyAnnotation]) -> Factor:
        this: Factor = self
        return Factor(this.Name, map(f, this.FactorType), this.Comments)

    def SetCategory(self, c: OntologyAnnotation) -> Factor:
        this: Factor = self
        return Factor(this.Name, c, this.Comments)

    @staticmethod
    def try_get_by_name(name: str, factors: FSharpList[Factor]) -> Factor | None:
        def _arrow742(f: Factor) -> bool:
            return equals(f.Name, name)

        return try_find(_arrow742, factors)

    @staticmethod
    def exists_by_name(name: str, factors: FSharpList[Factor]) -> bool:
        def _arrow743(f: Factor) -> bool:
            return equals(f.Name, name)

        return exists(_arrow743, factors)

    @staticmethod
    def add(factors: FSharpList[Factor], factor: Factor) -> FSharpList[Factor]:
        return append(factors, singleton(factor))

    @staticmethod
    def remove_by_name(name: str, factors: FSharpList[Factor]) -> FSharpList[Factor]:
        def _arrow744(f: Factor) -> bool:
            return not equals(f.Name, name)

        return filter(_arrow744, factors)

    @staticmethod
    def get_comments(factor: Factor) -> Array[Comment] | None:
        return factor.Comments

    @staticmethod
    def map_comments(f: Callable[[Array[Comment]], Array[Comment]], factor: Factor) -> Factor:
        return Factor(factor.Name, factor.FactorType, map(f, factor.Comments))

    @staticmethod
    def set_comments(factor: Factor, comments: Array[Comment]) -> Factor:
        return Factor(factor.Name, factor.FactorType, comments)

    @staticmethod
    def get_factor_type(factor: Factor) -> OntologyAnnotation | None:
        return factor.FactorType

    @staticmethod
    def map_factor_type(f: Callable[[OntologyAnnotation], OntologyAnnotation], factor: Factor) -> Factor:
        return Factor(factor.Name, map(f, factor.FactorType), factor.Comments)

    @staticmethod
    def set_factor_type(factor: Factor, factor_type: OntologyAnnotation) -> Factor:
        return Factor(factor.Name, factor_type, factor.Comments)

    @staticmethod
    def try_get_name(f: Factor) -> str | None:
        return f.Name

    @staticmethod
    def get_name_as_string(f: Factor) -> str:
        return f.NameText

    @staticmethod
    def name_equals_string(name: str, f: Factor) -> bool:
        return f.NameText == name

    def Copy(self, __unit: None=None) -> Factor:
        this: Factor = self
        def mapping(a: Array[Comment]) -> Array[Comment]:
            def f(c: Comment, a: Any=a) -> Comment:
                return c.Copy()

            return ResizeArray_map(f, a)

        comments: Array[Comment] | None = map(mapping, this.Comments)
        return Factor.make(this.Name, this.FactorType, comments)

    def Print(self, __unit: None=None) -> str:
        this: Factor = self
        return to_string(this)

    def PrintCompact(self, __unit: None=None) -> str:
        this: Factor = self
        return "OA " + this.NameText


Factor_reflection = _expr745

__all__ = ["Factor_reflection"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ...fable_library.array_ import insert_range_in_place
from ...fable_library.list import FSharpList
from ...fable_library.map import (of_seq as of_seq_1, try_find)
from ...fable_library.map_util import (add_to_dict, add_to_set)
from ...fable_library.option import (value as value_2, default_arg, map as map_2, bind)
from ...fable_library.reflection import (TypeInfo, class_type)
from ...fable_library.seq import (try_find_index, find_index, length, to_list, delay, map as map_1, choose, reduce, append)
from ...fable_library.seq2 import distinct
from ...fable_library.set import (intersect, of_seq, FSharpSet__get_IsEmpty)
from ...fable_library.types import Array
from ...fable_library.util import (get_enumerator, dispose, compare, string_hash, IEnumerable_1, compare_primitives, equals, ignore, IEnumerator, to_iterator)
from ..Helper.collections_ import (Dictionary_tryFind, ResizeArray_iter, ResizeArray_fold, ResizeArray_map, ResizeArray_groupBy, ResizeArray_collect)
from .arc_table import ArcTable
from .composite_cell import CompositeCell
from .composite_column import CompositeColumn
from .composite_header import (IOType, CompositeHeader)

def ArcTablesAux_tryFindIndexByTableName(name: str, tables: Array[ArcTable]) -> int | None:
    def predicate(t: ArcTable, name: Any=name, tables: Any=tables) -> bool:
        return t.Name == name

    return try_find_index(predicate, tables)


def ArcTablesAux_findIndexByTableName(name: str, tables: Array[ArcTable]) -> int:
    def predicate(t: ArcTable, name: Any=name, tables: Any=tables) -> bool:
        return t.Name == name

    match_value: int | None = try_find_index(predicate, tables)
    if match_value is None:
        raise Exception(("Unable to find table with name \'" + name) + "\'!")

    else: 
        return match_value



def ArcTablesAux_getIOMap(tables: Array[ArcTable]) -> Any:
    mappings: Any = dict([])
    def include_in_map(name: str, io_type: IOType, tables: Any=tables) -> None:
        if name != "":
            match_value: IOType | None = Dictionary_tryFind(name, mappings)
            if match_value is None:
                add_to_dict(mappings, name, io_type)

            else: 
                old_iotype: IOType = match_value
                new_iotype: IOType = old_iotype.Merge(io_type)
                mappings[name] = new_iotype



    enumerator: Any = get_enumerator(tables)
    try: 
        while enumerator.System_Collections_IEnumerator_MoveNext():
            table: ArcTable = enumerator.System_Collections_Generic_IEnumerator_1_get_Current()
            match_value_1: CompositeColumn | None = table.TryGetInputColumn()
            if match_value_1 is None:
                pass

            else: 
                ic: CompositeColumn = match_value_1
                io_type_1: IOType = value_2(ic.Header.TryInput())
                def f(c: CompositeCell) -> None:
                    include_in_map(c.ToFreeTextCell().AsFreeText, io_type_1)

                ResizeArray_iter(f, ic.Cells)

            match_value_2: CompositeColumn | None = table.TryGetOutputColumn()
            if match_value_2 is None:
                pass

            else: 
                oc: CompositeColumn = match_value_2
                io_type_2: IOType = value_2(oc.Header.TryOutput())
                def f_1(c_1: CompositeCell) -> None:
                    include_in_map(c_1.ToFreeTextCell().AsFreeText, io_type_2)

                ResizeArray_iter(f_1, oc.Cells)


    finally: 
        dispose(enumerator)

    return mappings


def ArcTablesAux_applyIOMap(map: Any, tables: Array[ArcTable]) -> None:
    enumerator: Any = get_enumerator(tables)
    try: 
        while enumerator.System_Collections_IEnumerator_MoveNext():
            table: ArcTable = enumerator.System_Collections_Generic_IEnumerator_1_get_Current()
            match_value: CompositeColumn | None = table.TryGetInputColumn()
            if match_value is None:
                pass

            else: 
                ic: CompositeColumn = match_value
                def predicate(x: CompositeHeader) -> bool:
                    return x.is_input

                index: int = find_index(predicate, table.Headers) or 0
                def f(io: IOType, c: CompositeCell) -> IOType:
                    match_value_1: IOType | None = Dictionary_tryFind(c.ToFreeTextCell().AsFreeText, map)
                    if match_value_1 is None:
                        return io

                    else: 
                        new_io: IOType = match_value_1
                        return io.Merge(new_io)


                new_iotype: IOType = ResizeArray_fold(f, value_2(ic.Header.TryInput()), ic.Cells)
                table.UpdateHeader(index, CompositeHeader(11, new_iotype))

            match_value_2: CompositeColumn | None = table.TryGetOutputColumn()
            if match_value_2 is None:
                pass

            else: 
                oc: CompositeColumn = match_value_2
                def predicate_1(x_1: CompositeHeader) -> bool:
                    return x_1.is_output

                index_1: int = find_index(predicate_1, table.Headers) or 0
                def f_1(io_1: IOType, c_1: CompositeCell) -> IOType:
                    match_value_3: IOType | None = Dictionary_tryFind(c_1.ToFreeTextCell().AsFreeText, map)
                    if match_value_3 is None:
                        return io_1

                    else: 
                        new_io_1: IOType = match_value_3
                        return io_1.Merge(new_io_1)


                new_iotype_1: IOType = ResizeArray_fold(f_1, value_2(oc.Header.TryOutput()), oc.Cells)
                table.UpdateHeader(index_1, CompositeHeader(12, new_iotype_1))


    finally: 
        dispose(enumerator)



def ArcTablesAux_SanityChecks_validateSheetIndex(index: int, allow_append: bool, sheets: Array[ArcTable]) -> None:
    if index < 0:
        raise Exception("Cannot insert ArcTable at index < 0.")

    def _arrow888(__unit: None=None, index: Any=index, allow_append: Any=allow_append, sheets: Any=sheets) -> bool:
        x: int = index or 0
        y: int = len(sheets) or 0
        return (compare(x, y) > 0) if allow_append else (compare(x, y) >= 0)

    if _arrow888():
        raise Exception(("Specified index is out of range! Assay contains only " + str(len(sheets))) + " tables.")



def ArcTablesAux_SanityChecks_validateNamesUnique(names: IEnumerable_1[str]) -> None:
    class ObjectExpr890:
        @property
        def Equals(self) -> Callable[[str, str], bool]:
            def _arrow889(x: str, y: str) -> bool:
                return x == y

            return _arrow889

        @property
        def GetHashCode(self) -> Callable[[str], int]:
            return string_hash

    if not (length(names) == length(distinct(names, ObjectExpr890()))):
        raise Exception("Cannot add multiple tables with the same name! Table names inside one assay must be unqiue")



def ArcTablesAux_SanityChecks_validateNewNameUnique(new_name: str, existing_names: IEnumerable_1[str]) -> None:
    def _arrow891(x: str, new_name: Any=new_name, existing_names: Any=existing_names) -> bool:
        return x == new_name

    match_value: int | None = try_find_index(_arrow891, existing_names)
    if match_value is None:
        pass

    else: 
        raise Exception(((("Cannot create table with name " + new_name) + ", as table names must be unique and table at index ") + str(match_value)) + " has the same name.")



def ArcTablesAux_SanityChecks_validateNewNameAtUnique(index: int, new_name: str, existing_names: IEnumerable_1[str]) -> None:
    def _arrow892(x: str, index: Any=index, new_name: Any=new_name, existing_names: Any=existing_names) -> bool:
        return x == new_name

    match_value: int | None = try_find_index(_arrow892, existing_names)
    if match_value is None:
        pass

    elif index == match_value:
        i_1: int = match_value or 0

    else: 
        i_2: int = match_value or 0
        raise Exception(((("Cannot create table with name " + new_name) + ", as table names must be unique and table at index ") + str(i_2)) + " has the same name.")



def ArcTablesAux_SanityChecks_validateNewNamesUnique(new_names: IEnumerable_1[str], existing_names: IEnumerable_1[str]) -> None:
    ArcTablesAux_SanityChecks_validateNamesUnique(new_names)
    class ObjectExpr893:
        @property
        def Compare(self) -> Callable[[str, str], int]:
            return compare_primitives

    class ObjectExpr894:
        @property
        def Compare(self) -> Callable[[str, str], int]:
            return compare_primitives

    same: Any = intersect(of_seq(new_names, ObjectExpr893()), of_seq(existing_names, ObjectExpr894()))
    if not FSharpSet__get_IsEmpty(same):
        raise Exception(("Cannot create tables with the names " + str(same)) + ", as table names must be unique.")



def _expr920() -> TypeInfo:
    return class_type("ARCtrl.ArcTables", None, ArcTables)


class ArcTables:
    def __init__(self, init_tables: Array[ArcTable]) -> None:
        def _expr919():
            def mapping(t: ArcTable) -> str:
                return t.Name

            ArcTablesAux_SanityChecks_validateNamesUnique(map_1(mapping, init_tables))
            return init_tables

        self.tables: Array[ArcTable] = _expr919()

    @property
    def Tables(self, __unit: None=None) -> Array[ArcTable]:
        this: ArcTables = self
        return this.tables

    @Tables.setter
    def Tables(self, new_tables: Array[ArcTable]) -> None:
        this: ArcTables = self
        this.tables = new_tables

    def get_item(self, index: int) -> ArcTable:
        this: ArcTables = self
        return this.Tables[index]

    @property
    def TableNames(self, __unit: None=None) -> FSharpList[str]:
        this: ArcTables = self
        def _arrow896(__unit: None=None) -> IEnumerable_1[str]:
            def _arrow895(s: ArcTable) -> str:
                return s.Name

            return map_1(_arrow895, this.Tables)

        return to_list(delay(_arrow896))

    @property
    def TableCount(self, __unit: None=None) -> int:
        this: ArcTables = self
        return len(this.Tables)

    def AddTable(self, table: ArcTable, index: int | None=None) -> None:
        this: ArcTables = self
        index_1: int = default_arg(index, this.TableCount) or 0
        ArcTablesAux_SanityChecks_validateSheetIndex(index_1, True, this.Tables)
        ArcTablesAux_SanityChecks_validateNewNameUnique(table.Name, this.TableNames)
        this.Tables.insert(index_1, table)

    def AddTables(self, tables: IEnumerable_1[ArcTable], index: int | None=None) -> None:
        this: ArcTables = self
        index_1: int = default_arg(index, this.TableCount) or 0
        ArcTablesAux_SanityChecks_validateSheetIndex(index_1, True, this.Tables)
        def mapping(x: ArcTable) -> str:
            return x.Name

        ArcTablesAux_SanityChecks_validateNewNamesUnique(map_1(mapping, tables), this.TableNames)
        insert_range_in_place(index_1, tables, this.Tables)

    def InitTable(self, table_name: str, index: int | None=None) -> ArcTable:
        this: ArcTables = self
        index_1: int = default_arg(index, this.TableCount) or 0
        table: ArcTable = ArcTable.init(table_name)
        ArcTablesAux_SanityChecks_validateSheetIndex(index_1, True, this.Tables)
        ArcTablesAux_SanityChecks_validateNewNameUnique(table.Name, this.TableNames)
        this.Tables.insert(index_1, table)
        return table

    def InitTables(self, table_names: IEnumerable_1[str], index: int | None=None) -> None:
        this: ArcTables = self
        index_1: int = default_arg(index, this.TableCount) or 0
        def mapping(x: str) -> ArcTable:
            return ArcTable.init(x)

        tables: IEnumerable_1[ArcTable] = map_1(mapping, table_names)
        ArcTablesAux_SanityChecks_validateSheetIndex(index_1, True, this.Tables)
        def mapping_1(x_1: ArcTable) -> str:
            return x_1.Name

        ArcTablesAux_SanityChecks_validateNewNamesUnique(map_1(mapping_1, tables), this.TableNames)
        insert_range_in_place(index_1, tables, this.Tables)

    def GetTableAt(self, index: int) -> ArcTable:
        this: ArcTables = self
        ArcTablesAux_SanityChecks_validateSheetIndex(index, False, this.Tables)
        return this.Tables[index]

    def GetTable(self, name: str) -> ArcTable:
        this: ArcTables = self
        index: int = ArcTablesAux_findIndexByTableName(name, this.Tables) or 0
        return this.GetTableAt(index)

    def UpdateTableAt(self, index: int, table: ArcTable) -> None:
        this: ArcTables = self
        ArcTablesAux_SanityChecks_validateSheetIndex(index, False, this.Tables)
        ArcTablesAux_SanityChecks_validateNewNameAtUnique(index, table.Name, this.TableNames)
        this.Tables[index] = table

    def UpdateTable(self, name: str, table: ArcTable) -> None:
        this: ArcTables = self
        tupled_arg: tuple[int, ArcTable] = (ArcTablesAux_findIndexByTableName(name, this.Tables), table)
        this.UpdateTableAt(tupled_arg[0], tupled_arg[1])

    def SetTableAt(self, index: int, table: ArcTable) -> None:
        this: ArcTables = self
        ArcTablesAux_SanityChecks_validateSheetIndex(index, True, this.Tables)
        ArcTablesAux_SanityChecks_validateNewNameAtUnique(index, table.Name, this.TableNames)
        this.Tables[index] = table

    def SetTable(self, name: str, table: ArcTable) -> None:
        this: ArcTables = self
        match_value: int | None = ArcTablesAux_tryFindIndexByTableName(name, this.Tables)
        if match_value is None:
            this.AddTable(table)

        else: 
            index: int = match_value or 0
            this.SetTableAt(index, table)


    def RemoveTableAt(self, index: int) -> None:
        this: ArcTables = self
        ArcTablesAux_SanityChecks_validateSheetIndex(index, False, this.Tables)
        this.Tables.pop(index)

    def RemoveTable(self, name: str) -> None:
        this: ArcTables = self
        index: int = ArcTablesAux_findIndexByTableName(name, this.Tables) or 0
        this.RemoveTableAt(index)

    def MapTableAt(self, index: int, update_fun: Callable[[ArcTable], None]) -> None:
        this: ArcTables = self
        ArcTablesAux_SanityChecks_validateSheetIndex(index, False, this.Tables)
        update_fun(this.Tables[index])

    def MapTable(self, name: str, update_fun: Callable[[ArcTable], None]) -> None:
        this: ArcTables = self
        tupled_arg: tuple[int, Callable[[ArcTable], None]] = (ArcTablesAux_findIndexByTableName(name, this.Tables), update_fun)
        this.MapTableAt(tupled_arg[0], tupled_arg[1])

    def RenameTableAt(self, index: int, new_name: str) -> None:
        this: ArcTables = self
        ArcTablesAux_SanityChecks_validateSheetIndex(index, False, this.Tables)
        ArcTablesAux_SanityChecks_validateNewNameUnique(new_name, this.TableNames)
        table: ArcTable = this.GetTableAt(index)
        table.Name = new_name

    def RenameTable(self, name: str, new_name: str) -> None:
        this: ArcTables = self
        tupled_arg: tuple[int, str] = (ArcTablesAux_findIndexByTableName(name, this.Tables), new_name)
        this.RenameTableAt(tupled_arg[0], tupled_arg[1])

    def AddColumnAt(self, table_index: int, header: CompositeHeader, cells: Array[CompositeCell] | None=None, column_index: int | None=None, force_replace: bool | None=None) -> None:
        this: ArcTables = self
        def _arrow899(table: ArcTable) -> None:
            table.AddColumn(header, cells, column_index, force_replace)

        this.MapTableAt(table_index, _arrow899)

    def AddColumn(self, table_name: str, header: CompositeHeader, cells: Array[CompositeCell] | None=None, column_index: int | None=None, force_replace: bool | None=None) -> None:
        this: ArcTables = self
        i: int = ArcTablesAux_findIndexByTableName(table_name, this.Tables) or 0
        this.AddColumnAt(i, header, cells, column_index, force_replace)

    def RemoveColumnAt(self, table_index: int, column_index: int) -> None:
        this: ArcTables = self
        def _arrow902(table: ArcTable) -> None:
            table.RemoveColumn(column_index)

        this.MapTableAt(table_index, _arrow902)

    def RemoveColumn(self, table_name: str, column_index: int) -> None:
        this: ArcTables = self
        tupled_arg: tuple[int, int] = (ArcTablesAux_findIndexByTableName(table_name, this.Tables), column_index)
        this.RemoveColumnAt(tupled_arg[0], tupled_arg[1])

    def UpdateColumnAt(self, table_index: int, column_index: int, header: CompositeHeader, cells: Array[CompositeCell] | None=None) -> None:
        this: ArcTables = self
        def _arrow905(table: ArcTable) -> None:
            table.UpdateColumn(column_index, header, cells)

        this.MapTableAt(table_index, _arrow905)

    def UpdateColumn(self, table_name: str, column_index: int, header: CompositeHeader, cells: Array[CompositeCell] | None=None) -> None:
        this: ArcTables = self
        table_index: int = ArcTablesAux_findIndexByTableName(table_name, this.Tables) or 0
        this.UpdateColumnAt(table_index, column_index, header, cells)

    def GetColumnAt(self, table_index: int, column_index: int) -> CompositeColumn:
        this: ArcTables = self
        table: ArcTable = this.GetTableAt(table_index)
        return table.GetColumn(column_index)

    def GetColumn(self, table_name: str, column_index: int) -> CompositeColumn:
        this: ArcTables = self
        tupled_arg: tuple[int, int] = (ArcTablesAux_findIndexByTableName(table_name, this.Tables), column_index)
        return this.GetColumnAt(tupled_arg[0], tupled_arg[1])

    def AddRowAt(self, table_index: int, cells: Array[CompositeCell] | None=None, row_index: int | None=None) -> None:
        this: ArcTables = self
        def _arrow908(table: ArcTable) -> None:
            table.AddRow(cells, row_index)

        this.MapTableAt(table_index, _arrow908)

    def AddRow(self, table_name: str, cells: Array[CompositeCell] | None=None, row_index: int | None=None) -> None:
        this: ArcTables = self
        i: int = ArcTablesAux_findIndexByTableName(table_name, this.Tables) or 0
        this.AddRowAt(i, cells, row_index)

    def RemoveRowAt(self, table_index: int, row_index: int) -> None:
        this: ArcTables = self
        def _arrow910(table: ArcTable) -> None:
            table.RemoveRow(row_index)

        this.MapTableAt(table_index, _arrow910)

    def RemoveRow(self, table_name: str, row_index: int) -> None:
        this: ArcTables = self
        tupled_arg: tuple[int, int] = (ArcTablesAux_findIndexByTableName(table_name, this.Tables), row_index)
        this.RemoveRowAt(tupled_arg[0], tupled_arg[1])

    def UpdateRowAt(self, table_index: int, row_index: int, cells: Array[CompositeCell]) -> None:
        this: ArcTables = self
        def _arrow912(table: ArcTable) -> None:
            table.UpdateRow(row_index, cells)

        this.MapTableAt(table_index, _arrow912)

    def UpdateRow(self, table_name: str, row_index: int, cells: Array[CompositeCell]) -> None:
        this: ArcTables = self
        tupled_arg: tuple[int, int, Array[CompositeCell]] = (ArcTablesAux_findIndexByTableName(table_name, this.Tables), row_index, cells)
        this.UpdateRowAt(tupled_arg[0], tupled_arg[1], tupled_arg[2])

    def GetRowAt(self, table_index: int, row_index: int) -> Array[CompositeCell]:
        this: ArcTables = self
        table: ArcTable = this.GetTableAt(table_index)
        return table.GetRow(row_index)

    def GetRow(self, table_name: str, row_index: int) -> Array[CompositeCell]:
        this: ArcTables = self
        tupled_arg: tuple[int, int] = (ArcTablesAux_findIndexByTableName(table_name, this.Tables), row_index)
        return this.GetRowAt(tupled_arg[0], tupled_arg[1])

    @staticmethod
    def of_seq(tables: IEnumerable_1[ArcTable]) -> ArcTables:
        return ArcTables(list(tables))

    def MoveTable(self, old_index: int, new_index: int) -> None:
        this: ArcTables = self
        table: ArcTable = this.GetTableAt(old_index)
        this.Tables.pop(old_index)
        this.Tables.insert(new_index, table)

    @staticmethod
    def update_reference_tables_by_sheets(reference_tables: ArcTables, sheet_tables: ArcTables, keep_unused_ref_tables: bool | None=None) -> ArcTables:
        keep_unused_ref_tables_1: bool = default_arg(keep_unused_ref_tables, False)
        used_tables: Any = set([])
        def chooser(t: ArcTable) -> tuple[str, ArcTable] | None:
            def mapping(c: CompositeColumn, t: Any=t) -> tuple[str, ArcTable]:
                return (c.Cells[0].AsFreeText, t)

            return map_2(mapping, t.TryGetProtocolNameColumn())

        class ObjectExpr915:
            @property
            def Compare(self) -> Callable[[str, str], int]:
                return compare_primitives

        reference_table_map: Any = of_seq_1(choose(chooser, reference_tables.Tables), ObjectExpr915())
        def _arrow917(__unit: None=None) -> IEnumerable_1[ArcTable]:
            def f_3(tupled_arg: tuple[str, Array[ArcTable]]) -> ArcTable:
                def reduction(table: ArcTable, table_1: ArcTable, tupled_arg: Any=tupled_arg) -> ArcTable:
                    return ArcTable.append(table, table_1)

                return reduce(reduction, tupled_arg[1])

            def f_2(t_2: ArcTable) -> str:
                return t_2.Name

            def f_1(t_1: ArcTable) -> ArcTable:
                def binder_1(c_1: CompositeCell, t_1: Any=t_1) -> str | None:
                    if c_1.AsFreeText == "":
                        return None

                    else: 
                        return c_1.AsFreeText


                def binder(i: int, t_1: Any=t_1) -> CompositeCell | None:
                    return t_1.TryGetCellAt(i, 0)

                def predicate(x_1: CompositeHeader, t_1: Any=t_1) -> bool:
                    return equals(x_1, CompositeHeader(8))

                k: str = default_arg(bind(binder_1, bind(binder, try_find_index(predicate, t_1.Headers))), t_1.Name)
                match_value: ArcTable | None = try_find(k, reference_table_map)
                if match_value is None:
                    return t_1

                else: 
                    rt: ArcTable = match_value
                    ignore(add_to_set(k, used_tables))
                    updated_table: ArcTable = ArcTable.update_reference_by_annotation_table(rt, t_1)
                    return ArcTable.from_arc_table_values(t_1.Name, updated_table.Headers, updated_table.Values)


            def _arrow916(__unit: None=None) -> Array[ArcTable]:
                a: Array[ArcTable] = sheet_tables.Tables
                return ResizeArray_collect(ArcTable.SplitByProtocolREF(), a)

            s: Array[ArcTable] = ResizeArray_map(f_3, ResizeArray_groupBy(f_2, ResizeArray_map(f_1, _arrow916())))
            def chooser_1(kv: Any) -> ArcTable | None:
                if kv[0] in used_tables:
                    return None

                else: 
                    return kv[1]


            return append(choose(chooser_1, reference_table_map), s) if keep_unused_ref_tables_1 else s

        return ArcTables(list(_arrow917()))

    def GetEnumerator(self, __unit: None=None) -> IEnumerator[ArcTable]:
        this: ArcTables = self
        return get_enumerator(this.Tables)

    def __iter__(self) -> IEnumerator[ArcTable]:
        return to_iterator(self.GetEnumerator())

    def System_Collections_IEnumerable_GetEnumerator(self, __unit: None=None) -> IEnumerator[Any]:
        this: ArcTables = self
        return get_enumerator(this.Tables)


ArcTables_reflection = _expr920

def ArcTables__ctor_Z420F2E1A(init_tables: Array[ArcTable]) -> ArcTables:
    return ArcTables(init_tables)


__all__ = ["ArcTablesAux_tryFindIndexByTableName", "ArcTablesAux_findIndexByTableName", "ArcTablesAux_getIOMap", "ArcTablesAux_applyIOMap", "ArcTablesAux_SanityChecks_validateSheetIndex", "ArcTablesAux_SanityChecks_validateNamesUnique", "ArcTablesAux_SanityChecks_validateNewNameUnique", "ArcTablesAux_SanityChecks_validateNewNameAtUnique", "ArcTablesAux_SanityChecks_validateNewNamesUnique", "ArcTables_reflection"]




from __future__ import annotations
from abc import abstractmethod
from datetime import datetime as datetime_1
from typing import (Protocol, Any)
from ..fable_library.date import to_universal_time
from ..fable_library.option import some
from ..fable_library.util import equals
from ..fable_openpyxl.openpyxl import Cell
from ..fs_spreadsheet.Cells.fs_cell import (DataType, FsCell)
from ..fs_spreadsheet.fs_address import FsAddress

class datetime(Protocol):
    @abstractmethod
    def decoy(self) -> None:
        ...


class DateTimeStatic(Protocol):
    @abstractmethod
    def from_time_stamp(self, timestamp: float) -> datetime:
        ...


def to_universal_time_py(dt: Any) -> datetime:
    timestamp: float = to_universal_time(dt).timestamp()
    return datetime_1.fromtimestamp(timestamp=timestamp)


def from_fs_cell(fs_cell: FsCell) -> Any | None:
    match_value: DataType = fs_cell.DataType
    if match_value.tag == 2:
        return some(fs_cell.ValueAsFloat())

    elif match_value.tag == 3:
        return some(to_universal_time_py(fs_cell.ValueAsDateTime()))

    elif match_value.tag == 0:
        return some(fs_cell.Value)

    elif match_value.tag == 4:
        return some(fs_cell.Value)

    else: 
        return some(fs_cell.ValueAsBool())



def to_fs_cell(worksheet_name: Any, row_index: int, column_index: int, py_cell: Cell) -> FsCell:
    fsadress: FsAddress = FsAddress(row_index, column_index)
    pattern_input_1: tuple[DataType, Any]
    pattern_input: tuple[DataType, Any]
    value_1: Any = py_cell.value
    pattern_input = ((DataType(0), value_1)) if (str(type(value_1)) == "<class \'str\'>") else ((((DataType(1), True)) if value_1 else ((DataType(1), False))) if (str(type(value_1)) == "<class \'bool\'>") else (((DataType(2), value_1)) if (str(type(value_1)) == "<fable_modules.fable_library.types.uint8\'>>") else (((DataType(2), value_1)) if (str(type(value_1)) == "<class \'fable_modules.fable_library.types.int8\'>") else (((DataType(2), value_1)) if (str(type(value_1)) == "<class \'int\'>") else (((DataType(2), value_1)) if (str(type(value_1)) == "<class \'fable_modules.fable_library.types.int16\'>") else (((DataType(2), value_1)) if (str(type(value_1)) == "<class \'fable_modules.fable_library.types.int64\'>") else (((DataType(2), value_1)) if (str(type(value_1)) == "<class \'fable_modules.fable_library.types.uint32>") else (((DataType(2), value_1)) if (str(type(value_1)) == "<class \'fable_modules.fable_library.types.uint16\'>") else (((DataType(2), value_1)) if (str(type(value_1)) == "<class \'fable_modules.fable_library.types.uint32\'>") else (((DataType(2), value_1)) if (str(type(value_1)) == "<class \'fable_modules.fable_library.types.float32\'>") else (((DataType(2), value_1)) if (str(type(value_1)) == "<class \'float\'>") else (((DataType(2), value_1)) if (str(type(value_1)) == "<class \'int\'>") else (((DataType(3), value_1)) if isinstance(value_1, datetime_1) else (((DataType(0), value_1)) if (str(type(value_1)) == "<class \'str\'>") else ((DataType(0), value_1))))))))))))))))
    v: Any = pattern_input[1]
    pattern_input_1 = ((DataType(1), True)) if (True if equals(v, "=TRUE()") else equals(v, "=True()")) else (((DataType(1), False)) if (True if equals(v, "=FALSE()") else equals(v, "=False()")) else ((pattern_input[0], v)))
    return FsCell(pattern_input_1[1], pattern_input_1[0], fsadress)


__all__ = ["to_universal_time_py", "from_fs_cell", "to_fs_cell"]

